import numpy as np
import copy

from qtpy.QtWidgets import (
    QPushButton,
    QHBoxLayout,
    QVBoxLayout,
    QCheckBox,
    QTableWidget,
    QWidget,
    QTableWidgetItem,
    QHeaderView,
    QMessageBox,
)
from qtpy.QtGui import QBrush, QColor, QDoubleValidator
from qtpy.QtCore import Qt, Slot, Signal

from .useful_widgets import (
    LineEditReadOnly,
    ElementSelection,
    SecondaryWindow,
    set_tooltip,
    LineEditExtended,
    CheckBoxNamed,
    get_background_css,
)

from .dlg_edit_user_peak_parameters import DialogEditUserPeakParameters
from .dlg_new_user_peak import DialogNewUserPeak
from .dlg_pileup_peak_parameters import DialogPileupPeakParameters

import logging

logger = logging.getLogger(__name__)


class WndManageEmissionLines(SecondaryWindow):

    signal_selected_element_changed = Signal(str)
    signal_update_element_selection_list = Signal()
    signal_update_add_remove_btn_state = Signal(bool, bool)
    signal_marker_state_changed = Signal(bool)

    signal_parameters_changed = Signal()

    def __init__(self, *, gpc, gui_vars):
        super().__init__()

        # Global processing classes
        self.gpc = gpc
        # Global GUI variables (used for control of GUI state)
        self.gui_vars = gui_vars

        # Threshold used for peak removal (displayed in lineedit)
        self._remove_peak_threshold = self.gpc.get_peak_threshold()

        self._enable_events = False

        self._eline_list = []  # List of emission lines (used in the line selection combo)
        self._table_contents = []  # Keep a copy of table contents (list of dict)
        self._selected_eline = ""

        self.initialize()

        self._enable_events = True

        # Marker state is reported by Matplotlib plot in 'line_plot' model
        def cb(marker_state):
            self.signal_marker_state_changed.emit(marker_state)

        self.gpc.set_marker_reporter(cb)
        self.signal_marker_state_changed.connect(self.slot_marker_state_changed)

        # Update button states
        self._update_add_remove_btn_state()
        self._update_add_edit_userpeak_btn_state()
        self._update_add_edit_pileup_peak_btn_state()

    def initialize(self):
        self.setWindowTitle("PyXRF: Add/Remove Emission Lines")
        self.resize(600, 600)

        top_buttons = self._setup_select_elines()
        self._setup_elines_table()
        bottom_buttons = self._setup_action_buttons()

        vbox = QVBoxLayout()

        # Group of buttons above the table
        vbox.addLayout(top_buttons)

        # Tables
        hbox = QHBoxLayout()
        hbox.addWidget(self.tbl_elines)
        vbox.addLayout(hbox)

        vbox.addLayout(bottom_buttons)

        self.setLayout(vbox)

        self._set_tooltips()

    def _setup_select_elines(self):

        self.cb_select_all = QCheckBox("All")
        self.cb_select_all.setChecked(True)
        self.cb_select_all.toggled.connect(self.cb_select_all_toggled)

        self.element_selection = ElementSelection()

        # The following field should switched to 'editable' state from when needed
        self.le_peak_intensity = LineEditReadOnly()

        self.pb_add_eline = QPushButton("Add")
        self.pb_add_eline.clicked.connect(self.pb_add_eline_clicked)

        self.pb_remove_eline = QPushButton("Remove")
        self.pb_remove_eline.clicked.connect(self.pb_remove_eline_clicked)

        self.pb_user_peaks = QPushButton("New User Peak ...")
        self.pb_user_peaks.clicked.connect(self.pb_user_peaks_clicked)
        self.pb_pileup_peaks = QPushButton("New Pileup Peak ...")
        self.pb_pileup_peaks.clicked.connect(self.pb_pileup_peaks_clicked)

        self.element_selection.signal_current_item_changed.connect(self.element_selection_item_changed)

        vbox = QVBoxLayout()

        hbox = QHBoxLayout()
        hbox.addWidget(self.element_selection)
        hbox.addWidget(self.le_peak_intensity)
        hbox.addWidget(self.pb_add_eline)
        hbox.addWidget(self.pb_remove_eline)
        vbox.addLayout(hbox)

        hbox = QHBoxLayout()
        hbox.addWidget(self.cb_select_all)
        hbox.addStretch(1)
        hbox.addWidget(self.pb_user_peaks)
        hbox.addWidget(self.pb_pileup_peaks)
        vbox.addLayout(hbox)

        # Wrap vbox into hbox, because it will be inserted into vbox
        hbox = QHBoxLayout()
        hbox.addLayout(vbox)
        hbox.addStretch(1)
        return hbox

    def _setup_elines_table(self):
        """The table has only functionality necessary to demonstrate how it is going
        to look. A lot more code is needed to actually make it run."""

        self._validator_peak_height = QDoubleValidator()
        self._validator_peak_height.setBottom(0.01)

        self.tbl_elines = QTableWidget()
        self.tbl_elines.setStyleSheet(
            "QTableWidget::item{color: black;}"
            "QTableWidget::item:selected{background-color: red;}"
            "QTableWidget::item:selected{color: white;}"
        )

        self.tbl_labels = ["", "Z", "Line", "E, keV", "Peak Int.", "Rel. Int.(%)", "CS"]
        self.tbl_cols_editable = ["Peak Int."]
        self.tbl_value_min = {"Rel. Int.(%)": 0.1}
        tbl_cols_resize_to_content = ["", "Z", "Line"]

        self.tbl_elines.setColumnCount(len(self.tbl_labels))
        self.tbl_elines.verticalHeader().hide()
        self.tbl_elines.setHorizontalHeaderLabels(self.tbl_labels)

        self.tbl_elines.setSelectionBehavior(QTableWidget.SelectRows)
        self.tbl_elines.setSelectionMode(QTableWidget.SingleSelection)
        self.tbl_elines.itemSelectionChanged.connect(self.tbl_elines_item_selection_changed)
        self.tbl_elines.itemChanged.connect(self.tbl_elines_item_changed)

        header = self.tbl_elines.horizontalHeader()
        for n, lbl in enumerate(self.tbl_labels):
            # Set stretching for the columns
            if lbl in tbl_cols_resize_to_content:
                header.setSectionResizeMode(n, QHeaderView.ResizeToContents)
            else:
                header.setSectionResizeMode(n, QHeaderView.Stretch)
            # Set alignment for the columns headers (HEADERS only)
            if n == 0:
                header.setDefaultAlignment(Qt.AlignCenter)
            else:
                header.setDefaultAlignment(Qt.AlignRight)

        self.cb_sel_list = []  # List of checkboxes

    def _setup_action_buttons(self):
        self.pb_remove_rel = QPushButton("Remove Rel.Int.(%) <")
        self.pb_remove_rel.clicked.connect(self.pb_remove_rel_clicked)

        self.le_remove_rel = LineEditExtended("")
        self._validator_le_remove_rel = QDoubleValidator()
        self._validator_le_remove_rel.setBottom(0.01)  # Some small number
        self._validator_le_remove_rel.setTop(100.0)
        self.le_remove_rel.setText(self._format_threshold(self._remove_peak_threshold))
        self._update_le_remove_rel_state()
        self.le_remove_rel.textChanged.connect(self.le_remove_rel_text_changed)
        self.le_remove_rel.editingFinished.connect(self.le_remove_rel_editing_finished)

        self.pb_remove_unchecked = QPushButton("Remove Unchecked Lines")
        self.pb_remove_unchecked.clicked.connect(self.pb_remove_unchecked_clicked)

        hbox = QHBoxLayout()
        hbox.addWidget(self.pb_remove_rel)
        hbox.addWidget(self.le_remove_rel)
        hbox.addStretch(1)
        hbox.addWidget(self.pb_remove_unchecked)

        return hbox

    def _set_tooltips(self):
        set_tooltip(self.cb_select_all, "<b>Select/Deselect All</b> emission lines in the list")
        set_tooltip(self.element_selection, "<b>Set active</b> emission line")
        set_tooltip(self.le_peak_intensity, "Set or modify <b>intensity</b> of the active peak.")
        set_tooltip(self.pb_add_eline, "<b>Add</b> emission line to the list.")
        set_tooltip(self.pb_remove_eline, "<b>Remove</b> emission line from the list.")
        set_tooltip(
            self.pb_user_peaks, "Open dialog box to add or modify parameters of the <b>user-defined peak</b>"
        )
        set_tooltip(self.pb_pileup_peaks, "Open dialog box to add or modify parameters of the <b>pileup peak</b>")

        set_tooltip(self.tbl_elines, "The list of the selected <b>emission lines</b>")

        # set_tooltip(self.pb_update,
        #             "Update the internally stored list of selected emission lines "
        #             "and their parameters. This button is <b>deprecated</b>, but still may be "
        #             "needed in some situations. In future releases it will be <b>removed</b> or replaced "
        #             "with 'Accept' button. Substantial changes to the computational code is needed before "
        #             "it happens.")
        # set_tooltip(self.pb_undo,
        #             "<b>Undo</b> changes to the table of selected emission lines. Doesn't always work.")
        set_tooltip(
            self.pb_remove_rel,
            "<b>Remove emission lines</b> from the list if their relative intensity is less "
            "then specified threshold.",
        )
        set_tooltip(
            self.le_remove_rel,
            "<b>Threshold</b> that controls which emission lines are removed "
            "when <b>Remove Rel.Int.(%)</b> button is pressed.",
        )
        set_tooltip(self.pb_remove_unchecked, "Remove <b>unchecked</b> emission lines from the list.")

    def update_widget_state(self, condition=None):
        # Update the state of the menu bar
        state = not self.gui_vars["gui_state"]["running_computations"]
        self.setEnabled(state)

        # Hide the window if required by the program state
        state_file_loaded = self.gui_vars["gui_state"]["state_file_loaded"]
        state_model_exist = self.gui_vars["gui_state"]["state_model_exists"]
        if not state_file_loaded or not state_model_exist:
            self.hide()

        if condition == "tooltips":
            self._set_tooltips()

    def fill_eline_table(self, table_contents):
        self._table_contents = copy.deepcopy(table_contents)

        self._enable_events = False

        self.tbl_elines.clearContents()

        # Clear the list of checkboxes
        for cb in self.cb_sel_list:
            cb.stateChanged.connect(self.cb_eline_state_changed)
        self.cb_sel_list = []

        self.tbl_elines.setRowCount(len(table_contents))
        for nr, row in enumerate(table_contents):
            sel_status = row["sel_status"]
            row_data = [None, row["z"], row["eline"], row["energy"], row["peak_int"], row["rel_int"], row["cs"]]

            for nc, entry in enumerate(row_data):

                label = self.tbl_labels[nc]

                # Set alternating background colors for the table rows
                #   Make background for editable items a little brighter
                brightness = 240 if label in self.tbl_cols_editable else 220
                if nr % 2:
                    rgb_bckg = (255, brightness, brightness)  # Light-red
                else:
                    rgb_bckg = (brightness, 255, brightness)  # Light-green

                if nc == 0:
                    item = QWidget()
                    cb = CheckBoxNamed(name=f"{nr}")
                    item_hbox = QHBoxLayout(item)
                    item_hbox.addWidget(cb)
                    item_hbox.setAlignment(Qt.AlignCenter)
                    item_hbox.setContentsMargins(0, 0, 0, 0)

                    css1 = get_background_css(rgb_bckg, widget="QCheckbox", editable=False)
                    css2 = get_background_css(rgb_bckg, widget="QWidget", editable=False)
                    item.setStyleSheet(css2 + css1)

                    cb.setChecked(Qt.Checked if sel_status else Qt.Unchecked)
                    cb.stateChanged.connect(self.cb_eline_state_changed)
                    cb.setStyleSheet("QCheckBox {color: black;}")
                    self.cb_sel_list.append(cb)

                    self.tbl_elines.setCellWidget(nr, nc, item)
                else:

                    s = None
                    # The case when the value (Rel. Int.) is limited from the bottom
                    #   We don't want to print very small numbers here
                    if label in self.tbl_value_min:
                        v = self.tbl_value_min[label]
                        if isinstance(entry, (float, np.float64)) and (entry < v):
                            s = f"<{v:.2f}"
                    if s is None:
                        if isinstance(entry, (float, np.float64)):
                            s = f"{entry:.2f}" if entry else "-"
                        else:
                            s = f"{entry}" if entry else "-"

                    item = QTableWidgetItem(s)

                    item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)

                    # Set all columns not editable (unless needed)
                    if label not in self.tbl_cols_editable:
                        item.setFlags(item.flags() & ~Qt.ItemIsEditable)

                    brush = QBrush(QColor(*rgb_bckg))
                    item.setBackground(brush)

                    self.tbl_elines.setItem(nr, nc, item)

        self._enable_events = True
        # Update the rest of the widgets
        self._update_widgets_based_on_table_state()

    @Slot()
    def update_widget_data(self):
        # This is typically a new set of emission lines. Clear the selection both
        #   in the table and in the element selection tool.
        self.element_selection.set_current_item("")
        self.tbl_elines.clearSelection()
        self._set_selected_eline("")
        # Now update the tables
        self._update_eline_selection_list()
        self.update_eline_table()
        self._update_add_remove_btn_state()

    def pb_pileup_peaks_clicked(self):
        data = {}

        eline = self._selected_eline
        if self.gpc.get_eline_name_category(eline) == "pileup":
            logger.error(f"Attempt to add pileup peak '{eline}' while another pileup peak is selected.")
            return

        energy, marker_visible = self.gpc.get_suggested_manual_peak_energy()
        best_guess = self.gpc.get_guessed_pileup_peak_components(energy=energy, tolerance=0.1)
        if best_guess is not None:
            el1, el2, energy = best_guess
        else:
            # No peaks were found, enter peaks manually
            el1, el2, energy = "", "", 0
        data["element1"] = el1
        data["element2"] = el2
        data["energy"] = energy
        data["range_low"], data["range_high"] = self.gpc.get_selected_energy_range()

        if not marker_visible:
            # We shouldn't end up here, but this will protect from crashing in case
            #   the button was not disabled (a bug).
            msg = "Select location of the new peak center (energy)\nby clicking on the plot in 'Fit Model' tab"
            msgbox = QMessageBox(QMessageBox.Information, "User Input Required", msg, QMessageBox.Ok, parent=self)
            msgbox.exec()
        else:
            dlg = DialogPileupPeakParameters()

            def func():
                def f(e1, e2):
                    try:
                        name = self.gpc.generate_pileup_peak_name(e1, e2)
                        e = self.gpc.get_pileup_peak_energy(name)
                    except Exception:
                        e = 0
                    return e

                return f

            dlg.set_compute_energy_function(func())
            dlg.set_parameters(data)
            if dlg.exec():
                print("Pileup peak is added")
                try:
                    data = dlg.get_parameters()
                    eline1, eline2 = data["element1"], data["element2"]
                    eline = self.gpc.generate_pileup_peak_name(eline1, eline2)
                    self.gpc.add_peak_manual(eline)
                    self.update_eline_table()  # Update the table
                    self.tbl_elines_set_selection(eline)  # Select new emission line
                    self._set_selected_eline(eline)
                    self._set_fit_status(False)
                    logger.info(f"New pileup peak {eline} was added")
                except RuntimeError as ex:
                    msg = str(ex)
                    msgbox = QMessageBox(QMessageBox.Critical, "Error", msg, QMessageBox.Ok, parent=self)
                    msgbox.exec()
                    # Reload the table anyway (nothing is going to be selected)
                    self.update_eline_table()

    def pb_user_peaks_clicked(self):
        eline = self._selected_eline
        # If current peak is user_defined peak
        is_userpeak = self.gpc.get_eline_name_category(eline) == "userpeak"

        if is_userpeak:
            data = {}
            data["enabled"] = True
            data["name"] = eline
            data["maxv"] = self.gpc.get_eline_intensity(eline)
            data["energy"], data["fwhm"] = self.gpc.get_current_userpeak_energy_fwhm()

            dlg = DialogEditUserPeakParameters()
            dlg.set_parameters(data=data)
            if dlg.exec():
                print("Editing of user defined peak is completed")
                try:
                    eline = data["name"]
                    data = dlg.get_parameters()
                    self.gpc.update_userpeak(data["name"], data["energy"], data["maxv"], data["fwhm"])
                    self._set_fit_status(False)
                    logger.info(f"User defined peak {eline} was updated.")
                except Exception as ex:
                    msg = str(ex)
                    msgbox = QMessageBox(QMessageBox.Critical, "Error", msg, QMessageBox.Ok, parent=self)
                    msgbox.exec()
                # Reload the table anyway (nothing is going to be selected)
                self.update_eline_table()

        else:
            data = {}
            data["name"] = self.gpc.get_unused_userpeak_name()
            data["energy"], marker_visible = self.gpc.get_suggested_manual_peak_energy()
            if marker_visible:
                dlg = DialogNewUserPeak()
                dlg.set_parameters(data=data)
                if dlg.exec():
                    try:
                        eline = data["name"]
                        self.gpc.add_peak_manual(eline)
                        self.update_eline_table()  # Update the table
                        self.tbl_elines_set_selection(eline)  # Select new emission line
                        self._set_selected_eline(eline)
                        self._set_fit_status(False)
                        logger.info(f"New user defined peak {eline} is added")
                    except RuntimeError as ex:
                        msg = str(ex)
                        msgbox = QMessageBox(QMessageBox.Critical, "Error", msg, QMessageBox.Ok, parent=self)
                        msgbox.exec()
                        # Reload the table anyway (nothing is going to be selected)
                        self.update_eline_table()
            else:
                msg = (
                    "Select location of the new peak center (energy)\n"
                    "by clicking on the plot in 'Fit Model' tab"
                )
                msgbox = QMessageBox(
                    QMessageBox.Information, "User Input Required", msg, QMessageBox.Ok, parent=self
                )
                msgbox.exec()

    @Slot()
    def pb_add_eline_clicked(self):
        logger.debug("'Add line' clicked")
        # It is assumed that this button is active only if an element is selected from the list
        #   of available emission lines. It can't be used to add user-defined peaks or pileup peaks.
        eline = self._selected_eline
        if eline:
            try:
                self.gpc.add_peak_manual(eline)
                self.update_eline_table()  # Update the table
                self.tbl_elines_set_selection(eline)  # Select new emission line
                self._set_fit_status(False)
            except RuntimeError as ex:
                msg = str(ex)
                msgbox = QMessageBox(QMessageBox.Critical, "Error", msg, QMessageBox.Ok, parent=self)
                msgbox.exec()
                # Reload the table anyway (nothing is going to be selected)
                self.update_eline_table()

    @Slot()
    def pb_remove_eline_clicked(self):
        logger.debug("'Remove line' clicked")
        eline = self._selected_eline
        if eline:
            # If currently selected line is the emission line (like Ca_K), we want
            # it to remain selected after it is deleted. This means that nothing is selected
            # in the table. For other lines, nothing should remain selected.
            self.tbl_elines.clearSelection()
            self.gpc.remove_peak_manual(eline)
            self.update_eline_table()  # Update the table
            if self.gpc.get_eline_name_category(eline) != "eline":
                eline = ""
            # This will update widgets
            self._set_selected_eline(eline)
            self._set_fit_status(False)

    def cb_select_all_toggled(self, state):
        self._enable_events = False

        eline_list, state_list = [], []

        for n_row in range(self.tbl_elines.rowCount()):
            eline = self._table_contents[n_row]["eline"]
            # Do not deselect lines in category 'other'. They probably never to be deleted.
            # They also could be deselected manually.
            if self.gpc.get_eline_name_category(eline) == "other" and not state:
                to_check = True
            else:
                to_check = state
            self.cb_sel_list[n_row].setChecked(Qt.Checked if to_check else Qt.Unchecked)
            eline_list.append(eline)
            state_list.append(to_check)

        self.gpc.set_checked_emission_lines(eline_list, state_list)
        self._set_fit_status(False)
        self._enable_events = True

    def cb_eline_state_changed(self, name, state):
        if self._enable_events:
            n_row = int(name)
            state = state == Qt.Checked
            eline = self._table_contents[n_row]["eline"]
            self.gpc.set_checked_emission_lines([eline], [state])
            self._set_fit_status(False)

    def tbl_elines_item_changed(self, item):
        if self._enable_events:
            n_row, n_col = self.tbl_elines.row(item), self.tbl_elines.column(item)
            # Value was changed
            if n_col == 4:
                text = item.text()
                eline = self._table_contents[n_row]["eline"]
                if self._validator_peak_height.validate(text, 0)[0] != QDoubleValidator.Acceptable:
                    val = self._table_contents[n_row]["peak_int"]
                    self._enable_events = False
                    item.setText(f"{val:.2f}")
                    self._enable_events = True
                    self._set_fit_status(False)
                else:
                    self.gpc.update_eline_peak_height(eline, float(text))
                    self.update_eline_table()

    def tbl_elines_item_selection_changed(self):
        sel_ranges = self.tbl_elines.selectedRanges()
        # The table is configured to have one or no selected ranges
        # 'Open' button should be enabled only if a range (row) is selected
        if sel_ranges:
            index = sel_ranges[0].topRow()
            eline = self._table_contents[index]["eline"]
            if self._enable_events:
                self._enable_events = False
                self._set_selected_eline(eline)
                self.element_selection.set_current_item(eline)
                self._enable_events = True

    def tbl_elines_set_selection(self, eline):
        """
        Select the row with emission line `eline` in the table. Deselect everything if
        the emission line does not exist.
        """
        index = self._get_eline_index_in_table(eline)
        self.tbl_elines.clearSelection()
        if index >= 0:
            self.tbl_elines.selectRow(index)

    def element_selection_item_changed(self, index, eline):
        self.signal_selected_element_changed.emit(eline)

        if self._enable_events:
            self._enable_events = False
            self._set_selected_eline(eline)
            self.tbl_elines_set_selection(eline)
            self._enable_events = True

    def pb_remove_rel_clicked(self):
        try:
            self.gpc.remove_peaks_below_threshold(self._remove_peak_threshold)
        except Exception as ex:
            msg = str(ex)
            msgbox = QMessageBox(QMessageBox.Critical, "Error", msg, QMessageBox.Ok, parent=self)
            msgbox.exec()
        self.update_eline_table()
        # Update the displayed estimated peak amplitude value 'le_peak_intensity'
        self._set_selected_eline(self._selected_eline)
        self._set_fit_status(False)

    def le_remove_rel_text_changed(self, text):
        self._update_le_remove_rel_state(text)

    def le_remove_rel_editing_finished(self):
        text = self.le_remove_rel.text()
        if self._validator_le_remove_rel.validate(text, 0)[0] == QDoubleValidator.Acceptable:
            self._remove_peak_threshold = float(text)
        else:
            self.le_remove_rel.setText(self._format_threshold(self._remove_peak_threshold))

    def pb_remove_unchecked_clicked(self):
        try:
            self.gpc.remove_unchecked_peaks()
        except Exception as ex:
            msg = str(ex)
            msgbox = QMessageBox(QMessageBox.Critical, "Error", msg, QMessageBox.Ok, parent=self)
            msgbox.exec()
        # Reload the table
        self.update_eline_table()
        # Update the displayed estimated peak amplitude value 'le_peak_intensity'
        self._set_selected_eline(self._selected_eline)
        self._set_fit_status(False)

    def _display_peak_intensity(self, eline):
        v = self.gpc.get_eline_intensity(eline)
        s = f"{v:.10g}" if v is not None else ""
        self.le_peak_intensity.setText(s)

    def _update_le_remove_rel_state(self, text=None):
        if text is None:
            text = self.le_remove_rel.text()
        state = self._validator_le_remove_rel.validate(text, 0)[0] == QDoubleValidator.Acceptable
        self.le_remove_rel.setValid(state)
        self.pb_remove_rel.setEnabled(state)

    @Slot(str)
    def slot_selection_item_changed(self, eline):
        self.element_selection.set_current_item(eline)

    @Slot(bool)
    def slot_marker_state_changed(self, state):
        # If userpeak is selected and plot is clicked (marker is set), then user
        #   should be allowed to add userpeak at a new location. So deselect the userpeak
        #   from the table (if it is selected)
        logger.debug(f"Vertical marker on the fit plot changed state to {state}.")
        if state:
            self._deselect_special_peak_in_table()
        # Now update state of all buttons
        self._update_add_remove_btn_state()
        self._update_add_edit_userpeak_btn_state()
        self._update_add_edit_pileup_peak_btn_state()

    def _format_threshold(self, value):
        return f"{value:.2f}"

    def _deselect_special_peak_in_table(self):
        """Deselect userpeak if a userpeak is selected"""
        if self.gpc.get_eline_name_category(self._selected_eline) in ("userpeak", "pileup"):
            # Clear all selections
            self.tbl_elines_set_selection("")
            self._set_selected_eline("")
            # We also want to show marker at the new position
            self.gpc.show_marker_at_current_position()

    def _update_widgets_based_on_table_state(self):
        index, eline = self._get_current_index_in_table()
        if index >= 0:
            # Selection exists. Update the state of element selection widget.
            self.element_selection.set_current_item(eline)
        else:
            # No selection, update the state based on element selection widget.
            eline = self._selected_eline
            self.tbl_elines_set_selection(eline)
        self._update_add_remove_btn_state(eline)
        self._update_add_edit_userpeak_btn_state()
        self._update_add_edit_pileup_peak_btn_state()

    def _update_eline_selection_list(self):
        self._eline_list = self.gpc.get_full_eline_list()
        self.element_selection.set_item_list(self._eline_list)
        self.signal_update_element_selection_list.emit()

    @Slot()
    def update_eline_table(self):
        """Update table of emission lines without changing anything else"""
        eline_table = self.gpc.get_selected_eline_table()
        self.fill_eline_table(eline_table)

    def _get_eline_index_in_table(self, eline):
        try:
            index = [_["eline"] for _ in self._table_contents].index(eline)
        except ValueError:
            index = -1
        return index

    def _get_eline_index_in_list(self, eline):
        try:
            index = self._eline_list.index(eline)
        except ValueError:
            index = -1
        return index

    def _get_current_index_in_table(self):
        sel_ranges = self.tbl_elines.selectedRanges()
        # The table is configured to have one or no selected ranges
        # 'Open' button should be enabled only if a range (row) is selected
        if sel_ranges:
            index = sel_ranges[0].topRow()
            eline = self._table_contents[index]["eline"]
        else:
            index, eline = -1, ""
        return index, eline

    def _get_current_index_in_list(self):
        index, eline = self.element_selection.get_current_item()
        return index, eline

    def _update_add_remove_btn_state(self, eline=None):
        if eline is None:
            index_in_table, eline = self._get_current_index_in_table()
            index_in_list, eline = self._get_current_index_in_list()
        else:
            index_in_table = self._get_eline_index_in_table(eline)
            index_in_list = self._get_eline_index_in_list(eline)
        add_enabled, remove_enabled = True, True
        if index_in_list < 0 and index_in_table < 0:
            add_enabled, remove_enabled = False, False
        else:
            if index_in_table >= 0:
                if self.gpc.get_eline_name_category(eline) != "other":
                    add_enabled = False
                else:
                    add_enabled, remove_enabled = False, False
            else:
                remove_enabled = False
        self.pb_add_eline.setEnabled(add_enabled)
        self.pb_remove_eline.setEnabled(remove_enabled)
        self.signal_update_add_remove_btn_state.emit(add_enabled, remove_enabled)

    def _update_add_edit_userpeak_btn_state(self):

        enabled = True
        add_peak = True
        if self.gpc.get_eline_name_category(self._selected_eline) == "userpeak":
            add_peak = False

        # Finally check if marker is set (you need it for adding peaks)
        _, marker_set = self.gpc.get_suggested_manual_peak_energy()

        if not marker_set and add_peak:
            enabled = False

        if add_peak:
            btn_text = "New User Peak ..."
        else:
            btn_text = "Edit User Peak ..."
        self.pb_user_peaks.setText(btn_text)
        self.pb_user_peaks.setEnabled(enabled)

    def _update_add_edit_pileup_peak_btn_state(self):

        enabled = True
        if self.gpc.get_eline_name_category(self._selected_eline) == "pileup":
            enabled = False

        # Finally check if marker is set (you need it for adding peaks)
        _, marker_set = self.gpc.get_suggested_manual_peak_energy()
        # Ignore set marker for userpeaks (marker is used to display location of userpeaks)
        if self.gpc.get_eline_name_category(self._selected_eline) == "userpeak":
            marker_set = False

        if not marker_set:
            enabled = False

        self.pb_pileup_peaks.setEnabled(enabled)

    def _set_selected_eline(self, eline):
        self._update_add_remove_btn_state(eline)
        if eline != self._selected_eline:
            self._selected_eline = eline
            self.gpc.set_selected_eline(eline)
            self._display_peak_intensity(eline)
        else:
            # Peak intensity may change in some circumstances, so renew the displayed value.
            self._display_peak_intensity(eline)
        # Update button states after 'self._selected_eline' is set
        self._update_add_edit_userpeak_btn_state()
        self._update_add_edit_pileup_peak_btn_state()

    def _set_fit_status(self, status):
        self.gui_vars["gui_state"]["state_model_fit_exists"] = status
        self.signal_parameters_changed.emit()



from qtpy.QtWidgets import QTabWidget, QScrollArea
from qtpy.QtCore import Slot

from .tab_wd_load_data import LoadDataWidget
from .tab_wd_model import ModelWidget
from .tab_wd_fit_maps import FitMapsWidget


class LeftPanel(QTabWidget):
    def __init__(self, *, gpc, gui_vars):
        super().__init__()

        # Global processing classes
        self.gpc = gpc
        # Global GUI variables (used for control of GUI state)
        self.gui_vars = gui_vars

        self.setTabPosition(QTabWidget.West)

        self.load_data_widget = LoadDataWidget(gpc=self.gpc, gui_vars=self.gui_vars)
        self.load_data_tab = QScrollArea()
        self.load_data_tab.setWidget(self.load_data_widget)
        self.addTab(self.load_data_tab, "Data")

        self.model_widget = ModelWidget(gpc=self.gpc, gui_vars=self.gui_vars)
        self.model_tab = QScrollArea()
        self.model_tab.setWidget(self.model_widget)
        self.addTab(self.model_tab, "Model")

        self.fit_maps_widget = FitMapsWidget(gpc=self.gpc, gui_vars=self.gui_vars)
        self.fit_maps_tab = QScrollArea()
        self.fit_maps_tab.setWidget(self.fit_maps_widget)
        self.addTab(self.fit_maps_tab, "Maps")

        self.update_widget_state()

    def update_widget_state(self, condition=None):
        # TODO: this function has to enable tabs and widgets based on the current program state
        state = not self.gui_vars["gui_state"]["running_computations"]
        for i in range(self.count()):
            if state or (i != self.currentIndex()):
                self.setTabEnabled(i, state)
            self.widget(i).setEnabled(state)

        # Propagate the function call downstream (since the actual tab widget is 'QScrollArea')
        self.load_data_widget.update_widget_state(condition)
        self.model_widget.update_widget_state(condition)
        self.fit_maps_widget.update_widget_state(condition)

    @Slot(bool)
    def slot_activate_load_data_tab(self):
        self.setCurrentWidget(self.load_data_tab)



from __future__ import absolute_import

import numpy as np

# from scipy.interpolate import interp1d, interp2d
import re

import math
import matplotlib
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
import matplotlib.ticker as mticker
from mpl_toolkits.axes_grid1 import ImageGrid
from atom.api import Atom, Str, observe, Typed, Int, List, Dict, Bool, Float

from ..core.utils import normalize_data_by_scaler, grid_interpolate

from ..core.quant_analysis import ParamQuantitativeAnalysis
from ..core.xrf_utils import check_if_eline_supported

import logging

logger = logging.getLogger(__name__)


class DrawImageAdvanced(Atom):
    """
    This class performs 2D image rendering, such as showing multiple
    2D fitting or roi images based on user's selection.

    Attributes
    ----------
    img_data : dict
        dict of 2D array
    fig : object
        matplotlib Figure
    file_name : str
    stat_dict : dict
        determine which image to show
    data_opt : int
        index to show which data is chosen to plot
    dict_to_plot : dict
        selected data dict to plot, i.e., fitting data or roi is selected
    scale_opt : str
        linear or log plot
    color_opt : str
        orange or gray plot
    scaler_norm_dict : dict
        scaler normalization data, from img_dict
    scaler_items : list
        keys of scaler_norm_dict
    scaler_name_index : int
        index to select on GUI level
    scaler_data : None or numpy
        selected scaler data
    x_pos : list
        define data range in horizontal direction
    y_pos : list
        define data range in vertical direction
    pixel_or_pos : int
        index to choose plot with pixel (== 0) or with positions (== 1)
    grid_interpolate: bool
        choose to interpolate 2D image in terms of x,y or not
    limit_dict : Dict
        save low and high limit for image scaling
    """

    # Reference to FileIOMOdel
    io_model = Typed(object)

    fig = Typed(Figure)
    stat_dict = Dict()  # The value (bool) determines if the map (name is the key) is shown
    data_opt = Int(0)
    dict_to_plot = Dict()
    items_previous_selected = List()

    scale_opt = Str("Linear")
    color_opt = Str("viridis")
    img_title = Str()

    scaler_norm_dict = Dict()
    scaler_items = List()
    scaler_name_index = Int()
    scaler_data = Typed(object)

    x_pos = List()
    y_pos = List()

    pixel_or_pos = Int(0)
    grid_interpolate = Bool(False)
    map_keys = List()  # Keys for the map in the selected dataset
    limit_dict = Dict()
    range_dict = Dict()
    scatter_show = Bool(False)
    name_not_scalable = List()

    # The value is synchronized with currently selected value of incident energy
    #   Should not be changed directly
    incident_energy = Float(12.0)

    # Variable that indicates whether quanitative normalization should be applied to data
    #   Associated with 'Quantitative' checkbox
    quantitative_normalization = Bool(False)

    # The following fields are used for storing parameters used for quantitative analysis
    param_quant_analysis = Typed(object)
    # Distance to sample for the scan that is currently processed, synchronized with the contents
    #   of the field in 'Quantitative Analysis' tab, may be used for computations
    #   Enaml 'FloatField' requires this variable to be a string, but the string will always
    #   represent valid 'float' number
    quant_distance_to_sample = Float(0.0)
    # The following fields are used to facilitate GUI operation, not for long-term data storage
    #   Don't use those fields for any computations outside the GUI controls
    quant_calibration_data = List()
    quant_calibration_settings = List()
    quant_active_emission_lines = List()
    quant_calibration_data_preview = Str()

    def __init__(self, *, io_model):
        self.io_model = io_model

        self.fig = plt.figure(figsize=(3, 2))
        matplotlib.rcParams["axes.formatter.useoffset"] = True

        # Do not apply scaler norm on following data
        self.name_not_scalable = [
            "r2_adjust",
            "r_factor",
            "alive",
            "dead",
            "elapsed_time",
            "scaler_alive",
            "i0_time",
            "time",
            "time_diff",
            "dwell_time",
        ]

        self.param_quant_analysis = ParamQuantitativeAnalysis()
        self.param_quant_analysis.set_experiment_distance_to_sample(distance_to_sample=0.0)
        self.param_quant_analysis.set_experiment_incident_energy(incident_energy=self.incident_energy)

    def img_dict_updated(self, change):
        """
        Observer function to be connected to the fileio model
        in the top-level gui.py startup

        Parameters
        ----------
        changed : bool
            True - 'io_model.img_dict` was updated, False - ignore
        """
        if change["value"]:
            self.select_dataset(self.io_model.img_dict_default_selected_item)
            self.init_plot_status()

    def init_plot_status(self):
        # init of scaler for normalization
        self.scaler_name_index = 0

        scaler_groups = [v for v in list(self.io_model.img_dict.keys()) if "scaler" in v]
        if len(scaler_groups) > 0:
            # self.scaler_group_name = scaler_groups[0]
            self.scaler_norm_dict = self.io_model.img_dict[scaler_groups[0]]
            # for GUI purpose only
            self.scaler_items = []
            self.scaler_items = list(self.scaler_norm_dict.keys())
            self.scaler_items.sort()
            self.scaler_data = None

        # init of pos values
        self.set_pixel_or_pos(0)

        if "positions" in self.io_model.img_dict:
            try:
                logger.debug(f"Position keys: {list(self.io_model.img_dict['positions'].keys())}")
                self.x_pos = list(self.io_model.img_dict["positions"]["x_pos"][0, :])
                self.y_pos = list(self.io_model.img_dict["positions"]["y_pos"][:, -1])
                # when we use imshow, the x and y start at lower left,
                # so flip y, we want y starts from top left
                self.y_pos.reverse()

            except KeyError:
                pass
        else:
            self.x_pos = []
            self.y_pos = []

        logger.info("Use previously selected items as default: {}".format(self.items_previous_selected))
        logger.debug(
            "The following groups are included for 2D image display: {}".format(self.io_model.img_dict_keys)
        )

        self.show_image()

    def reset_to_default(self):
        """Set variables to default values as initiated."""
        self.select_dataset(0)
        # init of scaler for normalization
        self.set_scaler_index(0)
        self.plot_deselect_all()

    def get_detector_channel_name(self, img_title=None):
        # Return the name of the selected detector channel ('sum', 'det1', 'det2' etc)
        #   The channel name is extracted from 'self.img_title' (selected in 'ElementMap' tab)
        img_title = self.img_title if (img_title is None) else img_title
        if img_title:
            if re.search(r"_det\d+_fit$", img_title):
                s = re.search(r"_det\d+_", img_title)[0]
                return s.strip("_")
            if re.search(r"_fit", img_title):
                return "sum"
        else:
            return None

    def _get_current_plot_item(self):
        """Get the key for the current plot item (use in dictionary 'img_dict')"""
        return self.io_model.img_dict_keys[self.data_opt - 1]

    def set_map_keys(self):
        """
        Create sorted list of map keys. The list starts with sorted sequence of emission lines,
        followed by the sorted list of scalers and other maps.
        """
        self.map_keys.clear()
        # The key to use with 'img_dict', the name of the current dataset.
        plot_item = self._get_current_plot_item()
        keys_unsorted = list(self.io_model.img_dict[plot_item].keys())
        if len(keys_unsorted) != len(set(keys_unsorted)):
            logger.warning(
                f"DrawImageAdvanced:set_map_keys(): repeated keys in the dictionary 'img_dict': {keys_unsorted}"
            )
        keys_elines, keys_scalers = [], []
        for key in keys_unsorted:
            if check_if_eline_supported(key):  # Check if 'key' is an emission line (such as "Ca_K")
                keys_elines.append(key)
            else:
                keys_scalers.append(key)
        keys_elines.sort()
        keys_scalers.sort()
        self.map_keys = keys_elines + keys_scalers

    def select_dataset(self, dataset_index):
        """
        Select dataset. Meaning of the index: 0 - no dataset is selected,
        1, 2, ... datasets with index 0, 1, ... is selected

        Parameters
        ----------
        dataset_index: int
            index of the selected dataset
        """
        self.data_opt = dataset_index

        try:
            if self.data_opt == 0:
                self.dict_to_plot = {}
                self.map_keys.clear()
                self.set_stat_for_all(bool_val=False)
                self.img_title = ""
            elif self.data_opt > 0:
                # self.set_stat_for_all(bool_val=False)
                plot_item = self._get_current_plot_item()
                self.img_title = str(plot_item)
                self.dict_to_plot = self.io_model.img_dict[plot_item]
                self.set_map_keys()
                self.set_stat_for_all(bool_val=False)
                # Select the first item from the dataset (looks much better than empty plot)
                if self.map_keys:
                    self.stat_dict[self.map_keys[0]] = True

                logger.info(f"Dataset '{plot_item}' is selected.")

            # The detector channel name should be updated in any case
            self.param_quant_analysis.experiment_detector_channel = self.get_detector_channel_name()
            # Redraw image
            self.show_image()

        except IndexError:
            pass

    def get_selected_scaler_name(self):
        if self.scaler_name_index == 0:
            return None
        else:
            return self.scaler_items[self.scaler_name_index - 1]

    def set_scaler_index(self, scaler_index):

        self.scaler_name_index = scaler_index

        if self.scaler_name_index == 0:
            self.scaler_data = None
        else:
            try:
                scaler_name = self.scaler_items[self.scaler_name_index - 1]
            except IndexError:
                scaler_name = None
            if scaler_name:
                self.scaler_data = self.scaler_norm_dict[scaler_name]
                logger.info(
                    "Use scaler data to normalize, "
                    "and the shape of scaler data is {}, "
                    "with (low, high) as ({}, {})".format(
                        self.scaler_data.shape, np.min(self.scaler_data), np.max(self.scaler_data)
                    )
                )
        self.set_low_high_value()  # reset low high values based on normalization
        self.show_image()

    # TODO: document the following functions
    def update_quant_calibration_gui(self):
        self.quant_calibration_data = []
        self.quant_calibration_data = self.param_quant_analysis.calibration_data
        self.quant_calibration_settings = []
        self.quant_calibration_settings = self.param_quant_analysis.calibration_settings
        self.quant_active_emission_lines = []
        self.quant_active_emission_lines = self.param_quant_analysis.active_emission_lines

    def load_quantitative_calibration_data(self, file_path):
        self.param_quant_analysis.load_entry(file_path)
        self.update_quant_calibration_gui()

    def remove_quantitative_calibration_data(self, file_path):
        try:
            self.param_quant_analysis.remove_entry(file_path)
            self.update_quant_calibration_gui()
        except Exception as ex:
            logger.error(f"Calibration data was not removed: {ex}")

    def format_img_wizard_limit(self, value):
        """
        This function is used for formatting of range values in 'Image Wizard'.
        The presentation of the number was tweaked so that it is nicely formatted
           in the enaml field with adequate precision.

        ..note::

        The function is called externally from 'enaml' code.

        Parameters:
        ===========
        value : float
            The value to be formatted

        Returns:
        ========
        str - the string representation of the floating point variable
        """
        if value != 0:
            value_log10 = math.log10(abs(value))
        else:
            value_log10 = 0
        if (value_log10 > 3) or (value_log10 < -3):
            return f"{value:.6e}"
        return f"{value:.6f}"

    def set_scale_opt(self, scale_opt):
        scale_opt_list = ["Linear", "Log"]
        if scale_opt in scale_opt_list:
            self.scale_opt = scale_opt
            self.show_image()
        else:
            logger.error(f"XRF Maps: 'scale_opt' has illegal value: {scale_opt}. Allowed values: {scale_opt_list}")

    def set_color_opt(self, color_opt):
        self.color_opt = color_opt
        self.show_image()

    def set_pixel_or_pos(self, pixel_or_pos):
        self.pixel_or_pos = pixel_or_pos
        self.show_image()

    def set_grid_interpolate(self, grid_interpolate):
        self.grid_interpolate = grid_interpolate
        self.show_image()

    def enable_quantitative_normalization(self, enable):
        """
        Enable/Disable quantitative normalization.

        Parameters
        ----------
        enable: bool
            Enable quantitative normalization if True, disable if False.
        """
        self.quantitative_normalization = bool(enable)

        # Propagate current value of 'self.param_quant_analysis' (activate 'observer' functions)
        # TODO: the following may not be not needed in the view of the current framework
        tmp = self.param_quant_analysis
        self.param_quant_analysis = ParamQuantitativeAnalysis()
        self.param_quant_analysis = tmp

        self.set_low_high_value()  # reset low high values based on normalization
        self.show_image()

    def set_plot_scatter(self, is_scatter):
        self.scatter_show = is_scatter
        self.show_image()

    @observe("quant_distance_to_sample")
    def _on_change_distance_to_sample(self, change):
        # Set the value of the quantitative analysis parameter
        self.param_quant_analysis.set_experiment_distance_to_sample(self.quant_distance_to_sample)
        # Recompute range of plotted values (for each emission line in the dataset)
        self.set_low_high_value()

    def set_incident_energy(self, change):
        """
        The observer function that changes the value of incident energy
        and upper bound for fitted energy range. Should not be called directly.

        Parameters
        ----------

        change : dict
            ``change["value"]`` is the new value of incident energy
        """
        self.incident_energy = change["value"]

    @observe("incident_energy")
    def _on_change_incident_energy(self, change):
        self.param_quant_analysis.set_experiment_incident_energy(incident_energy=self.incident_energy)

    def set_stat_for_all(self, bool_val=False):
        """
        Set plotting status for all the 2D images, including low and high values.
        Note: 'self.map_keys' must be updated before calling this function!
        """
        self.stat_dict.clear()
        self.stat_dict = {k: bool_val for k in self.map_keys}

        self.limit_dict.clear()
        self.limit_dict = {k: {"low": 0.0, "high": 100.0} for k in self.map_keys}

        self.set_low_high_value()

    def set_low_high_value(self):
        """Set default low and high values based on normalization for each image."""
        # do not apply scaler norm on not scalable data
        self.range_dict.clear()

        for data_name in self.dict_to_plot.keys():

            if self.quantitative_normalization:
                # Quantitative normalization
                data_arr, _ = self.param_quant_analysis.apply_quantitative_normalization(
                    data_in=self.dict_to_plot[data_name],
                    scaler_dict=self.scaler_norm_dict,
                    scaler_name_default=self.get_selected_scaler_name(),
                    data_name=data_name,
                    name_not_scalable=self.name_not_scalable,
                )
            else:
                # Normalize by the selected scaler in a regular way
                data_arr = normalize_data_by_scaler(
                    data_in=self.dict_to_plot[data_name],
                    scaler=self.scaler_data,
                    data_name=data_name,
                    name_not_scalable=self.name_not_scalable,
                )

            lowv, highv = np.min(data_arr), np.max(data_arr)
            # Create some 'artificially' small range in case the array is constant
            if lowv == highv:
                lowv -= 0.005
                highv += 0.005
            self.range_dict[data_name] = {"low": lowv, "low_default": lowv, "high": highv, "high_default": highv}

    def reset_low_high(self, name):
        """Reset low and high value to default based on normalization."""
        self.range_dict[name]["low"] = self.range_dict[name]["low_default"]
        self.range_dict[name]["high"] = self.range_dict[name]["high_default"]
        self.limit_dict[name]["low"] = 0.0
        self.limit_dict[name]["high"] = 100.0
        self.show_image()

    def show_image(self):
        # Don't plot the image if dictionary is empty (causes a lot of issues)
        if not self.io_model.img_dict:
            return

        self.fig.clf()

        # The sequence of keys of 'img_dict' selected for plotting and ordered as 'map_keys'
        selected_keys = self.get_selected_items_for_plot()

        # Check if positions data is available. Positions data may be unavailable
        # (not recorded in HDF5 file) if experiment is has not been completed.
        # While the data from the completed part of experiment may still be used,
        # plotting vs. x-y or scatter plot may not be displayed.
        positions_data_available = False
        if "positions" in self.io_model.img_dict.keys():
            positions_data_available = True

        # Create local copies of self.pixel_or_pos, self.scatter_show and self.grid_interpolate
        pixel_or_pos_local = self.pixel_or_pos
        scatter_show_local = self.scatter_show
        grid_interpolate_local = self.grid_interpolate

        # Disable plotting vs x-y coordinates if 'positions' data is not available
        if not positions_data_available:
            if pixel_or_pos_local:
                pixel_or_pos_local = 0  # Switch to plotting vs. pixel number
                logger.error("'Positions' data is not available. Plotting vs. x-y coordinates is disabled")
            if scatter_show_local:
                scatter_show_local = False  # Switch to plotting vs. pixel number
                logger.error("'Positions' data is not available. Scatter plot is disabled.")
            if grid_interpolate_local:
                grid_interpolate_local = False  # Switch to plotting vs. pixel number
                logger.error("'Positions' data is not available. Interpolation is disabled.")

        rel_low_lim = 1e-6  # define the relative low limit for log image
        plot_interp = "Nearest"

        if self.scaler_data is not None:
            if np.count_nonzero(self.scaler_data) == 0:
                logger.warning("scaler is zero - scaling was not applied")
            elif len(self.scaler_data[self.scaler_data == 0]) > 0:
                logger.warning("scaler data has zero values")

        grey_use = self.color_opt

        ncol = int(np.ceil(np.sqrt(len(selected_keys))))
        try:
            nrow = int(np.ceil(len(selected_keys) / float(ncol)))
        except ZeroDivisionError:
            ncol = 1
            nrow = 1

        a_pad_v = 0.8
        a_pad_h = 0.5

        grid = ImageGrid(
            self.fig,
            111,
            nrows_ncols=(nrow, ncol),
            axes_pad=(a_pad_v, a_pad_h),
            cbar_location="right",
            cbar_mode="each",
            cbar_size="7%",
            cbar_pad="2%",
            share_all=True,
        )

        def _compute_equal_axes_ranges(x_min, x_max, y_min, y_max):
            """
            Compute ranges for x- and y- axes of the plot. Make sure that the ranges for x- and y-axes are
            always equal and fit the maximum of the ranges for x and y values:
                  max(abs(x_max-x_min), abs(y_max-y_min))
            The ranges are set so that the data is always centered in the middle of the ranges

            Parameters
            ----------

            x_min, x_max, y_min, y_max : float
                lower and upper boundaries of the x and y values

            Returns
            -------

            x_axis_min, x_axis_max, y_axis_min, y_axis_max : float
                lower and upper boundaries of the x- and y-axes ranges
            """

            x_axis_min, x_axis_max, y_axis_min, y_axis_max = x_min, x_max, y_min, y_max
            x_range, y_range = abs(x_max - x_min), abs(y_max - y_min)
            if x_range > y_range:
                y_center = (y_max + y_min) / 2
                y_axis_max = y_center + x_range / 2
                y_axis_min = y_center - x_range / 2
            else:
                x_center = (x_max + x_min) / 2
                x_axis_max = x_center + y_range / 2
                x_axis_min = x_center - y_range / 2

            return x_axis_min, x_axis_max, y_axis_min, y_axis_max

        def _adjust_data_range_using_min_ratio(c_min, c_max, c_axis_range, *, min_ratio=0.01):
            """
            Adjust the range for plotted data along one axis (x or y). The adjusted range is
            applied to the 'extend' attribute of imshow(). The adjusted range is always greater
            than 'axis_range * min_ratio'. Such transformation has no physical meaning
            and performed for aesthetic reasons: stretching the image presentation of
            a scan with only a few lines (1-3) greatly improves visibility of data.

            Parameters
            ----------

            c_min, c_max : float
                boundaries of the data range (along x or y axis)
            c_axis_range : float
                range presented along the same axis

            Returns
            -------

            cmin, c_max : float
                adjusted boundaries of the data range
            """
            c_range = c_max - c_min
            if c_range < c_axis_range * min_ratio:
                c_center = (c_max + c_min) / 2
                c_new_range = c_axis_range * min_ratio
                c_min = c_center - c_new_range / 2
                c_max = c_center + c_new_range / 2
            return c_min, c_max

        # Hide the axes that are unused (they are unsightly)
        for i in range(len(selected_keys), ncol * nrow):
            grid[i].set_visible(False)
            grid.cbar_axes[i].set_visible(False)

        for i, k in enumerate(selected_keys):

            quant_norm_applied = False
            if self.quantitative_normalization:
                # Quantitative normalization
                data_arr, quant_norm_applied = self.param_quant_analysis.apply_quantitative_normalization(
                    data_in=self.dict_to_plot[k],
                    scaler_dict=self.scaler_norm_dict,
                    scaler_name_default=self.get_selected_scaler_name(),
                    data_name=k,
                    name_not_scalable=self.name_not_scalable,
                )
            else:
                # Normalize by the selected scaler in a regular way
                data_arr = normalize_data_by_scaler(
                    data_in=self.dict_to_plot[k],
                    scaler=self.scaler_data,
                    data_name=k,
                    name_not_scalable=self.name_not_scalable,
                )

            if pixel_or_pos_local or scatter_show_local:

                # xd_min, xd_max, yd_min, yd_max = min(self.x_pos), max(self.x_pos),
                #     min(self.y_pos), max(self.y_pos)
                x_pos_2D = self.io_model.img_dict["positions"]["x_pos"]
                y_pos_2D = self.io_model.img_dict["positions"]["y_pos"]
                xd_min, xd_max, yd_min, yd_max = x_pos_2D.min(), x_pos_2D.max(), y_pos_2D.min(), y_pos_2D.max()
                xd_axis_min, xd_axis_max, yd_axis_min, yd_axis_max = _compute_equal_axes_ranges(
                    xd_min, xd_max, yd_min, yd_max
                )

                xd_min, xd_max = _adjust_data_range_using_min_ratio(xd_min, xd_max, xd_axis_max - xd_axis_min)
                yd_min, yd_max = _adjust_data_range_using_min_ratio(yd_min, yd_max, yd_axis_max - yd_axis_min)

                # Adjust the direction of each axis depending on the direction in which encoder values changed
                #   during the experiment. Data is plotted starting from the upper-right corner of the plot
                if x_pos_2D[0, 0] > x_pos_2D[0, -1]:
                    xd_min, xd_max, xd_axis_min, xd_axis_max = xd_max, xd_min, xd_axis_max, xd_axis_min
                if y_pos_2D[0, 0] > y_pos_2D[-1, 0]:
                    yd_min, yd_max, yd_axis_min, yd_axis_max = yd_max, yd_min, yd_axis_max, yd_axis_min

            else:

                yd, xd = data_arr.shape

                xd_min, xd_max, yd_min, yd_max = 0, xd, 0, yd
                if (yd <= math.floor(xd / 100)) and (xd >= 200):
                    yd_min, yd_max = -math.floor(xd / 200), math.ceil(xd / 200)
                if (xd <= math.floor(yd / 100)) and (yd >= 200):
                    xd_min, xd_max = -math.floor(yd / 200), math.ceil(yd / 200)

                xd_axis_min, xd_axis_max, yd_axis_min, yd_axis_max = _compute_equal_axes_ranges(
                    xd_min, xd_max, yd_min, yd_max
                )

            # Compute limits (used both for linear and log plots
            low_ratio = self.limit_dict[k]["low"] / 100.0
            high_ratio = self.limit_dict[k]["high"] / 100.0
            if (self.scaler_data is None) and (not quant_norm_applied):
                minv = self.range_dict[k]["low"]
                maxv = self.range_dict[k]["high"]
            else:
                # Unfortunately, the new normalization procedure requires to recalculate min and max values
                minv = np.min(data_arr)
                maxv = np.max(data_arr)
            low_limit = (maxv - minv) * low_ratio + minv
            high_limit = (maxv - minv) * high_ratio + minv
            # Set some minimum range for the colorbar (otherwise it will have white fill)
            if math.isclose(low_limit, high_limit, abs_tol=2e-20):
                if abs(low_limit) < 1e-20:  # The value is zero
                    dv = 1e-20
                else:
                    dv = math.fabs(low_limit * 0.01)
                high_limit += dv
                low_limit -= dv

            if self.scale_opt == "Linear":

                if not scatter_show_local:
                    if grid_interpolate_local:
                        data_arr, _, _ = grid_interpolate(
                            data_arr,
                            self.io_model.img_dict["positions"]["x_pos"],
                            self.io_model.img_dict["positions"]["y_pos"],
                        )
                    im = grid[i].imshow(
                        data_arr,
                        cmap=grey_use,
                        interpolation=plot_interp,
                        extent=(xd_min, xd_max, yd_max, yd_min),
                        origin="upper",
                        clim=(low_limit, high_limit),
                    )
                    grid[i].set_ylim(yd_axis_max, yd_axis_min)
                else:
                    xx = self.io_model.img_dict["positions"]["x_pos"]
                    yy = self.io_model.img_dict["positions"]["y_pos"]

                    # The following condition prevents crash if different file is loaded while
                    #    the scatter plot is open (PyXRF specific issue)
                    if data_arr.shape == xx.shape and data_arr.shape == yy.shape:
                        im = grid[i].scatter(
                            xx,
                            yy,
                            c=data_arr,
                            marker="s",
                            s=500,
                            alpha=1.0,  # Originally: alpha=0.8
                            cmap=grey_use,
                            vmin=low_limit,
                            vmax=high_limit,
                            linewidths=1,
                            linewidth=0,
                        )
                        grid[i].set_ylim(yd_axis_max, yd_axis_min)

                grid[i].set_xlim(xd_axis_min, xd_axis_max)

                grid_title = k
                if quant_norm_applied:
                    grid_title += " - Q"  # Mark the plots that represent quantitative information
                grid[i].text(0, 1.01, grid_title, ha="left", va="bottom", transform=grid[i].axes.transAxes)

                grid.cbar_axes[i].colorbar(im)
                im.colorbar.formatter = im.colorbar.ax.yaxis.get_major_formatter()

                # im.colorbar.ax.get_xaxis().set_ticks([])
                # im.colorbar.ax.get_xaxis().set_ticks([], minor=True)
                grid.cbar_axes[i].ticklabel_format(style="sci", scilimits=(-3, 4), axis="both")

                #  Do not remove this code, may be useful in the future (Dmitri G.) !!!
                #  Print label for colorbar
                # cax = grid.cbar_axes[i]
                # axis = cax.axis[cax.orientation]
                # axis.label.set_text("$[a.u.]$")

            else:

                # maxz = np.max(data_arr)
                # Set some reasonable minimum range for the colorbar
                #   Zeros or negative numbers will be shown in white
                # if maxz <= 1e-30:
                #     maxz = 1

                # Minimize from below for log scale
                if low_limit < high_limit * rel_low_lim:
                    low_limit = high_limit * rel_low_lim

                if not scatter_show_local:
                    if grid_interpolate_local:
                        data_arr, _, _ = grid_interpolate(
                            data_arr,
                            self.io_model.img_dict["positions"]["x_pos"],
                            self.io_model.img_dict["positions"]["y_pos"],
                        )
                    im = grid[i].imshow(
                        data_arr,
                        # norm=LogNorm(vmin=low_lim*maxz,
                        #              vmax=maxz, clip=True),
                        norm=LogNorm(vmin=low_limit, vmax=high_limit, clip=True),
                        cmap=grey_use,
                        interpolation=plot_interp,
                        extent=(xd_min, xd_max, yd_max, yd_min),
                        origin="upper",
                        # clim=(low_lim*maxz, maxz))
                        clim=(low_limit, high_limit),
                    )
                    grid[i].set_ylim(yd_axis_max, yd_axis_min)
                else:
                    im = grid[i].scatter(
                        self.io_model.img_dict["positions"]["x_pos"],
                        self.io_model.img_dict["positions"]["y_pos"],
                        # norm=LogNorm(vmin=low_lim*maxz,
                        #              vmax=maxz, clip=True),
                        norm=LogNorm(vmin=low_limit, vmax=high_limit, clip=True),
                        c=data_arr,
                        marker="s",
                        s=500,
                        alpha=1.0,  # Originally: alpha=0.8
                        cmap=grey_use,
                        linewidths=1,
                        linewidth=0,
                    )
                    grid[i].set_ylim(yd_axis_min, yd_axis_max)

                grid[i].set_xlim(xd_axis_min, xd_axis_max)

                grid_title = k
                if quant_norm_applied:
                    grid_title += " - Q"  # Mark the plots that represent quantitative information
                grid[i].text(0, 1.01, grid_title, ha="left", va="bottom", transform=grid[i].axes.transAxes)

                grid.cbar_axes[i].colorbar(im)
                im.colorbar.formatter = im.colorbar.ax.yaxis.get_major_formatter()
                im.colorbar.ax.get_xaxis().set_ticks([])
                im.colorbar.ax.get_xaxis().set_ticks([], minor=True)
                im.colorbar.ax.yaxis.set_minor_formatter(mticker.LogFormatter())

            grid[i].get_xaxis().set_major_locator(mticker.MaxNLocator(nbins="auto"))
            grid[i].get_yaxis().set_major_locator(mticker.MaxNLocator(nbins="auto"))

            grid[i].get_xaxis().get_major_formatter().set_useOffset(False)
            grid[i].get_yaxis().get_major_formatter().set_useOffset(False)

        # self.fig.suptitle(self.img_title, fontsize=20)

        self.fig.canvas.draw_idle()

    def get_selected_items_for_plot(self):
        """Collect the selected items for plotting."""
        # We want the dictionary to be sorted the same way as 'map_keys'
        sdict = self.stat_dict
        selected_keys = [_ for _ in self.map_keys if (_ in sdict) and (sdict[_] is True)]
        return selected_keys



import h5py

try:
    from databroker.v0 import Broker
except ModuleNotFoundError:
    from databroker import Broker

from databroker._core import register_builtin_handlers

#  srx detector, to be moved to filestore
# from databroker.assets.handlers import Xspress3HDF5Handler
from databroker.assets.handlers import HandlerBase

import logging

logger = logging.getLogger(__name__)

db = Broker.named("srx")
try:
    register_builtin_handlers(db.reg)
except Exception as ex:
    logger.error(f"Error while registering default SRX handlers: {ex}")


class BulkXSPRESS(HandlerBase):
    HANDLER_NAME = "XPS3_FLY"

    def __init__(self, resource_fn):
        self._handle = h5py.File(resource_fn, "r")

    def __call__(self):
        return self._handle["entry/instrument/detector/data"][:]


class ZebraHDF5Handler(HandlerBase):
    HANDLER_NAME = "ZEBRA_HDF51"

    def __init__(self, resource_fn):
        self._handle = h5py.File(resource_fn, "r")

    def __call__(self, column):
        return self._handle[column][:]


class SISHDF5Handler(HandlerBase):
    HANDLER_NAME = "SIS_HDF51"

    def __init__(self, resource_fn):
        self._handle = h5py.File(resource_fn, "r")

    def __call__(self, column):
        return self._handle[column][:]


class BulkMerlin(BulkXSPRESS):
    HANDLER_NAME = "MERLIN_FLY_STREAM_V1"

    def __call__(self):
        return self._handle["entry/instrument/detector/data"][:]


class BulkDexela(HandlerBase):
    HANDLER_NAME = "DEXELA_FLY_V1"

    def __init__(self, resource_fn):
        self._handle = h5py.File(resource_fn, "r")

    def __call__(self):
        return self._handle["entry/instrument/detector/data"][:]


db.reg.register_handler(BulkXSPRESS.HANDLER_NAME, BulkXSPRESS, overwrite=True)
db.reg.register_handler("SIS_HDF51", SISHDF5Handler, overwrite=True)
db.reg.register_handler("ZEBRA_HDF51", ZebraHDF5Handler, overwrite=True)
db.reg.register_handler(BulkMerlin.HANDLER_NAME, BulkMerlin, overwrite=True)
db.reg.register_handler(BulkDexela.HANDLER_NAME, BulkDexela, overwrite=True)



import jsonschema
import numpy as np
import pytest
import os

from pyxrf.core.yaml_param_files import (
    _parse_docstring_parameters,
    _verify_parsed_docstring,
    create_yaml_parameter_file,
    read_yaml_parameter_file,
)


def _generate_parameter_set():
    """
    Generates sample parameter dictionary for testing parameter manipulation functions
    """
    param_dict = {
        "param_none": None,
        "param_int": 265,
        "param_float": 345.834,
        "param_str1": "This_is_string_1",
        "param_str2": "This is string 2",
        "param_list_int": [2, 4, 6, 8, 10],
        "param_list_float": [2.2, 4.4, 6.6, 8.8, 10.01, 345623.453654762342, 3.453456e-15],
        "param_list_string": ["str1", "str2", "str3", "str4", "str5"],
        "param_list_misc": [2, 3.6, "str2", None, 3.45e10],
        "param_dictionary": {"p1": 65, "p2": 3.45, "p3": None, "p4": "some_string", "p5": [1, 2, 3.5]},
    }
    return param_dict


def _generate_parameter_set_schema():
    """
    Generates schema for the parameter set and validates the parameter set
    """

    param_schema = {
        "type": "object",
        "additionalProperties": False,
        "required": [
            "param_none",
            "param_int",
            "param_float",
            "param_str1",
            "param_str2",
            "param_list_int",
            "param_list_float",
            "param_list_string",
            "param_list_misc",
            "param_dictionary",
        ],
        "properties": {
            "param_none": {"type": "null"},
            "param_int": {"type": "integer"},
            "param_float": {"type": "number"},
            "param_str1": {"type": "string"},
            "param_str2": {"type": "string"},
            "param_list_int": {"type": "array", "items": {"type": "integer"}},
            "param_list_float": {"type": "array", "items": {"type": "number"}},
            "param_list_string": {"type": "array", "items": {"type": "string"}},
            "param_list_misc": {
                "type": "array",
                "additionalItems": False,
                "uniqueItems": True,
                "prefixItems": [
                    {"type": "integer"},
                    {"type": "number"},
                    {"type": "string"},
                    {"type": "null"},
                    {"type": "number"},
                ],
            },
            "param_dictionary": {
                "type": "object",
                "properties": {
                    "p1": {"type": "integer"},
                    "p2": {"type": "number"},
                    "p3": {"type": "null"},
                    "p4": {"type": "string"},
                    "p5": {"type": "array", "items": {"type": "number"}},
                },
            },
        },
    }

    # Make sure that the schema matches the parameter set
    param_dict = _generate_parameter_set()
    jsonschema.validate(instance=param_dict, schema=param_schema)

    return param_schema


def _generate_sample_docstring(param_dict, include_section_titles=True):
    """
    Generates sample docstring based on the supplied parameter dictionary,
    returns docstring and the list of parameter/description pairs.
    For testing of ``_parse_docstring_parameters`` function.
    If ``include_section_titles`` is ``False``, then the titles
    ``Parameters`` and ``Returns`` are not included in the output string,
    i.e. the string contains only parameter descriptions.
    """
    parameters = []
    for p_name in param_dict.keys():
        # Create tuple (param_name, param_description), param_description - array of strings,
        #   each string will be printed in the separate line
        p = (
            f"{p_name}",
            [
                f"{p_name} : {type(param_dict[p_name])}",
                f"Description of parameter {p_name}",
                "",
                f"The end of the description of {p_name}",
            ],
        )
        parameters.append(p)

    n_empty_lines_before, n_empty_lines_after = 5, 5
    d_str = [""] * n_empty_lines_before

    if include_section_titles:
        d_str.append("    Parameters")
        d_str.append("    ----------")

    for p in parameters:
        # Indentation by 4 spaces
        st = [f"    {s}" if s else s for s in p[1]]
        s = "\n".join(st)
        s += "\n" * np.random.choice(4)  # Insert 0 .. 3 empty strings
        d_str.append(s)

    if include_section_titles:
        d_str.append("    Returns")
        d_str.append("    -------")

    d_str.extend([""] * n_empty_lines_after)

    d_str = "\n".join(d_str)  # Convert the list to a single string

    return d_str, parameters


def test_parse_docstring_parameters():
    # Simple test for the successfully parsed docstring. It seems sufficient, since all error cases are trivial.

    param_dict = _generate_parameter_set()
    d_str, parameters = _generate_sample_docstring(param_dict)
    parameters_output = _parse_docstring_parameters(d_str)
    assert parameters == parameters_output, "Parsed parameters or parameter descriptions are invalid"

    # Basically repeats the same test, but enables section titles explicitely
    param_dict = _generate_parameter_set()
    d_str, parameters = _generate_sample_docstring(param_dict, include_section_titles=True)
    parameters_output = _parse_docstring_parameters(d_str, search_param_section=True)
    assert parameters == parameters_output, "Parsed parameters or parameter descriptions are invalid"

    # Disable section titles and see if parsing still works
    param_dict = _generate_parameter_set()
    d_str, parameters = _generate_sample_docstring(param_dict, include_section_titles=False)
    parameters_output = _parse_docstring_parameters(d_str, search_param_section=False)
    assert parameters == parameters_output, "Parsed parameters or parameter descriptions are invalid"

    # Check for exception if the section titles are required, but don't exist
    param_dict = _generate_parameter_set()
    d_str, parameters = _generate_sample_docstring(param_dict, include_section_titles=False)
    with pytest.raises(AssertionError, match="'Parameters' or 'Return' statement was not found in the docstring"):
        _parse_docstring_parameters(d_str, search_param_section=True)


def test_verify_parsed_docstring():

    # Generate the set of parameters (we don't use docstring in this test)
    param_dict = _generate_parameter_set()
    _, parameters = _generate_sample_docstring(param_dict)

    # This verification should be successful
    parameters_copy = parameters.copy()
    param_dict_copy = param_dict.copy()
    _verify_parsed_docstring(parameters, param_dict)  # This may raise an exception
    assert parameters == parameters_copy, "'parameters' was unintentionally changed by the function"
    assert param_dict == param_dict_copy, "'param_dict' was unintentionally changed by the function"

    # This test should fail (2 extra parameters)
    param_dict2 = param_dict.copy()
    param_dict2["extra_parameter1"] = 0
    param_dict2["extra_parameter2"] = 0

    parameters_copy = parameters.copy()
    param_dict_copy = param_dict2.copy()
    with pytest.raises(AssertionError, match="not found in the docstring.+extra_parameter1.+extra_parameter2"):
        _verify_parsed_docstring(parameters, param_dict2)
    assert parameters == parameters_copy, "'parameters' was unintentionally changed by the function"
    assert param_dict2 == param_dict_copy, "'param_dict' was unintentionally changed by the function"

    # This test should fail (1 parameter is removed)
    param_dict2 = param_dict.copy()
    # Select random key for removal
    key_to_remove = list(param_dict2.keys())[np.random.choice(len(param_dict2))]
    del param_dict2[key_to_remove]

    parameters_copy = parameters.copy()
    param_dict_copy = param_dict2.copy()
    with pytest.raises(AssertionError, match=f"not in the dictionary.+{key_to_remove}"):
        _verify_parsed_docstring(parameters, param_dict2)
    assert parameters == parameters_copy, "'parameters' was unintentionally changed by the function"
    assert param_dict2 == param_dict_copy, "'param_dict' was unintentionally changed by the function"


def test_create_read_yaml_parameter_file(tmp_path):

    # Some directory
    yaml_dirs = ["yaml", "file", "dirs"]
    yaml_fln = "parameter.yaml"
    file_path = os.path.join(tmp_path, *yaml_dirs, yaml_fln)

    # Generate the set of parameters (we don't use docstring in this test)
    param_dict = _generate_parameter_set()
    doc_string, parameters = _generate_sample_docstring(param_dict)

    create_yaml_parameter_file(
        file_path=file_path, function_docstring=doc_string, param_value_dict=param_dict, dir_create=True
    )

    param_dict_recovered = read_yaml_parameter_file(file_path=file_path)

    # Validate the schema of the recovered data (this will be part of the procedure of reading real data)
    param_schema = _generate_parameter_set_schema()
    jsonschema.validate(instance=param_dict_recovered, schema=param_schema)

    assert (
        param_dict == param_dict_recovered
    ), "Parameter dictionary read from YAML file is different from the original parameter dictionary"

    # Test: creating parameter file without instructions (and read it)
    yaml_fln2 = "parameter2.yaml"
    file_path2 = os.path.join(tmp_path, *yaml_dirs, yaml_fln2)
    create_yaml_parameter_file(
        file_path=file_path2,
        function_docstring=doc_string,
        param_value_dict=param_dict,
        dir_create=False,
        user_editing_instructions=False,
    )
    param_dict_recovered2 = read_yaml_parameter_file(file_path=file_path2)
    jsonschema.validate(instance=param_dict_recovered2, schema=param_schema)
    assert (
        param_dict == param_dict_recovered2
    ), "Parameter dictionary read from YAML file is different from the original parameter dictionary"

    # Test: creating parameter file with custom instructions (and read it)
    yaml_fln3 = "parameter3.yaml"
    file_path3 = os.path.join(tmp_path, *yaml_dirs, yaml_fln3)
    instructions = "    Those are custom instructions\n        in the correct format"
    create_yaml_parameter_file(
        file_path=file_path3,
        function_docstring=doc_string,
        param_value_dict=param_dict,
        dir_create=False,
        user_editing_instructions=instructions,
    )
    param_dict_recovered3 = read_yaml_parameter_file(file_path=file_path3)
    jsonschema.validate(instance=param_dict_recovered3, schema=param_schema)
    assert (
        param_dict == param_dict_recovered3
    ), "Parameter dictionary read from YAML file is different from the original parameter dictionary"


# fmt: off
@pytest.mark.parametrize("test_params", [
    # Writing to existing directory (absolute and relative paths)
    {"path_create": ["yaml", "file", "dirs"], "path_current": ["yaml", "file"],
     "path_write": ["."], "path_read": ["."], "dir_create": False,
     "succeed_write": True, "succeed_read": True},
    {"path_create": ["yaml", "file", "dirs"], "path_current": ["yaml", "file"],
     "path_write": [".", "dirs"], "path_read": ["yaml", "file", "dirs"],
     "dir_create": False, "succeed_write": True, "succeed_read": True},
    {"path_create": ["yaml", "file", "dirs"], "path_current": ["yaml", "file"],
     "path_write": ["yaml", "file", "dirs"], "path_read": [".", "dirs"],
     "dir_create": False, "succeed_write": True, "succeed_read": True},
    {"path_create": ["yaml", "file", "dirs"], "path_current": ["yaml", "file"],
     "path_write": [".."], "path_read": [".."], "dir_create": False,
     "succeed_write": True, "succeed_read": True},
    {"path_create": ["yaml", "file", "dirs"], "path_current": ["yaml", "file"],
     "path_write": [".."], "path_read": ["yaml"], "dir_create": False,
     "succeed_write": True, "succeed_read": True},
    {"path_create": ["yaml", "file", "dirs"], "path_current": ["yaml", "file"],
     "path_write": ["yaml"], "path_read": [".."], "dir_create": False,
     "succeed_write": True, "succeed_read": True},

    # Writing to non-existing directory
    {"path_create": ["yaml", "file"], "path_current": ["yaml", "file"],
     "path_write": [".", "dirs"], "path_read": [".", "dirs"], "dir_create": False,
     "succeed_write": False, "succeed_read": False},
    {"path_create": ["yaml", "file"], "path_current": ["yaml", "file"],
     "path_write": [".", "dirs"], "path_read": [".", "dirs"], "dir_create": True,
     "succeed_write": True, "succeed_read": True},

    # Reading non-existing file
    {"path_create": ["yaml", "file"], "path_current": ["yaml", "file"],
     "path_write": ["yaml"], "path_read": ["yaml", "file"], "dir_create": False,
     "succeed_write": True, "succeed_read": False},
])
# fmt: on
def test_create_yaml_parameter_file1(tmp_path, test_params):
    """Test functions ``create_yaml_parameter_file`` and ``read_yaml_parameter_file``
    for different combinations of input/output directories"""

    path_create = test_params["path_create"]
    path_current = test_params["path_current"]
    path_write = test_params["path_write"]
    path_read = test_params["path_read"]
    dir_create = test_params["dir_create"]
    succeed_write = test_params["succeed_write"]
    succeed_read = test_params["succeed_read"]

    def assemble_path(tmp_path, components):
        path = tmp_path
        if components:
            if components[0] in (".", ".."):
                # Relative path
                if components[0] == ".":
                    components.pop(0)
                if components:
                    path = os.path.join(*components)
                else:
                    path = ""
            else:
                # Absolute path
                components = [tmp_path] + components
                path = os.path.join(*components)
        return path

    path_create = assemble_path(tmp_path, path_create)
    path_current = assemble_path(tmp_path, path_current)
    path_write = assemble_path(tmp_path, path_write)
    path_read = assemble_path(tmp_path, path_read)

    # Generate the set of parameters (we don't use docstring in this test)
    param_dict = _generate_parameter_set()
    doc_string, parameters = _generate_sample_docstring(param_dict)

    os.makedirs(path_create, exist_ok=True)
    os.chdir(path_current)

    yaml_fln = "parameters.yaml"
    if succeed_write:
        create_yaml_parameter_file(
            file_path=os.path.join(path_write, yaml_fln),
            function_docstring=doc_string,
            param_value_dict=param_dict,
            dir_create=dir_create,
        )
    else:
        with pytest.raises(IOError):
            create_yaml_parameter_file(
                file_path=os.path.join(path_write, yaml_fln),
                function_docstring=doc_string,
                param_value_dict=param_dict,
                dir_create=dir_create,
            )

    if succeed_read:
        param_dict_recovered = read_yaml_parameter_file(file_path=os.path.join(path_read, yaml_fln))
    else:
        with pytest.raises(IOError):
            param_dict_recovered = read_yaml_parameter_file(file_path=os.path.join(path_read, yaml_fln))

    if succeed_write and succeed_read:
        assert (
            param_dict == param_dict_recovered
        ), "Parameter dictionary read from YAML file is different from the original parameter dictionary"


def test_create_yaml_parameter_file2(tmp_path):
    """Test if file overwriting protection works"""

    fln = os.path.join(tmp_path, "parameters.yaml")

    # Generate the set of parameters (we don't use docstring in this test)
    param_dict = _generate_parameter_set()
    doc_string, parameters = _generate_sample_docstring(param_dict)

    # Write the file the first time
    create_yaml_parameter_file(file_path=fln, function_docstring=doc_string, param_value_dict=param_dict)
    # Try to write it again
    with pytest.raises(IOError):
        create_yaml_parameter_file(file_path=fln, function_docstring=doc_string, param_value_dict=param_dict)
    # Try again, but allow overwriting the file
    create_yaml_parameter_file(
        file_path=fln, function_docstring=doc_string, param_value_dict=param_dict, file_overwrite=True
    )



from qtpy.QtWidgets import (
    QPushButton,
    QHBoxLayout,
    QVBoxLayout,
    QCheckBox,
    QLabel,
    QTableWidget,
    QTableWidgetItem,
    QHeaderView,
    QWidget,
    QMessageBox,
)
from qtpy.QtGui import QBrush, QColor
from qtpy.QtCore import Qt, Slot, Signal, QThreadPool, QRunnable

from .useful_widgets import (
    SecondaryWindow,
    set_tooltip,
    LineEditExtended,
    PushButtonNamed,
    CheckBoxNamed,
    RangeManager,
)

import logging

logger = logging.getLogger(__name__)


class WndComputeRoiMaps(SecondaryWindow):

    # Signal that is sent (to main window) to update global state of the program
    update_global_state = Signal()
    computations_complete = Signal(object)

    signal_roi_computation_complete = Signal()
    signal_activate_tab_xrf_maps = Signal()

    def __init__(self, *, gpc, gui_vars):
        super().__init__()

        # Global processing classes
        self.gpc = gpc
        # Global GUI variables (used for control of GUI state)
        self.gui_vars = gui_vars

        # Reference to the main window. The main window will hold
        #   references to all non-modal windows that could be opened
        #   from multiple places in the program.
        self.ref_main_window = self.gui_vars["ref_main_window"]

        self.update_global_state.connect(self.ref_main_window.update_widget_state)

        self.initialize()

    def initialize(self):
        self.setWindowTitle("PyXRF: Compute XRF Maps Based on ROIs")

        self.setMinimumWidth(600)
        self.setMinimumHeight(300)
        self.resize(600, 600)

        header_vbox = self._setup_header()
        self._setup_table()
        footer_hbox = self._setup_footer()

        vbox = QVBoxLayout()
        vbox.addLayout(header_vbox)
        vbox.addWidget(self.table)
        vbox.addLayout(footer_hbox)

        self.setLayout(vbox)

        self._set_tooltips()

    def _setup_header(self):
        self.pb_clear = QPushButton("Clear")
        self.pb_clear.clicked.connect(self.pb_clear_clicked)
        self.pb_use_lines_for_fitting = QPushButton("Use Lines Selected For Fitting")
        self.pb_use_lines_for_fitting.clicked.connect(self.pb_use_lines_for_fitting_clicked)

        self.le_sel_emission_lines = LineEditExtended()
        self.le_sel_emission_lines.textChanged.connect(self.le_sel_emission_lines_text_changed)
        self.le_sel_emission_lines.editingFinished.connect(self.le_sel_emission_lines_editing_finished)

        sample_elements = ""
        self.le_sel_emission_lines.setText(sample_elements)

        vbox = QVBoxLayout()
        hbox = QHBoxLayout()
        hbox.addWidget(QLabel("Enter emission lines, e.g. Fe_K, Gd_L  "))
        hbox.addStretch(1)
        hbox.addWidget(self.pb_clear)
        hbox.addWidget(self.pb_use_lines_for_fitting)
        vbox.addLayout(hbox)
        vbox.addWidget(self.le_sel_emission_lines)

        return vbox

    def _setup_table(self):

        # Labels for horizontal header
        self.tbl_labels = ["Line", "E, keV", "ROI, keV", "Show", "Reset"]

        # The list of columns that stretch with the table
        self.tbl_cols_stretch = ("E, keV", "ROI, keV")

        # Table item representation if different from default
        self.tbl_format = {"E, keV": ".3f"}

        # Editable items (highlighted with lighter background)
        self.tbl_cols_editable = {"ROI, keV"}

        # Columns that contain Range Manager
        self.tbl_cols_range_manager = ("ROI, keV",)

        self.table = QTableWidget()
        self.table.setColumnCount(len(self.tbl_labels))
        self.table.setHorizontalHeaderLabels(self.tbl_labels)
        self.table.verticalHeader().hide()
        self.table.setSelectionMode(QTableWidget.NoSelection)

        self.table.setStyleSheet("QTableWidget::item{color: black;}")

        header = self.table.horizontalHeader()
        for n, lbl in enumerate(self.tbl_labels):
            # Set stretching for the columns
            if lbl in self.tbl_cols_stretch:
                header.setSectionResizeMode(n, QHeaderView.Stretch)
            else:
                header.setSectionResizeMode(n, QHeaderView.ResizeToContents)

        self._table_contents = []
        self.cb_list = []
        self.range_manager_list = []
        self.pb_default_list = []
        self.fill_table(self._table_contents)

    def fill_table(self, table_contents):

        self.table.clearContents()
        self._table_contents = table_contents  # Save new table contents

        for item in self.range_manager_list:
            item.selection_changed.disconnect(self.range_manager_selection_changed)
        self.range_manager_list = []

        for cb in self.cb_list:
            cb.stateChanged.disconnect(self.cb_state_changed)
        self.cb_list = []

        for pb in self.pb_default_list:
            pb.clicked.connect(self.pb_default_clicked)
        self.pb_default_list = []

        self.table.setRowCount(len(table_contents))
        for nr, row in enumerate(table_contents):
            eline_name = row["eline"] + "a1"
            energy = row["energy_center"]
            energy_left = row["energy_left"]
            energy_right = row["energy_right"]
            range_displayed = row["range_displayed"]
            table_row = [eline_name, energy, (energy_left, energy_right)]
            for nc, entry in enumerate(table_row):

                label = self.tbl_labels[nc]

                # Set alternating background colors for the table rows
                #   Make background for editable items a little brighter
                brightness = 240 if label in self.tbl_cols_editable else 220
                if nr % 2:
                    rgb_bckg = (255, brightness, brightness)
                else:
                    rgb_bckg = (brightness, 255, brightness)

                if self.tbl_labels[nc] not in self.tbl_cols_range_manager:
                    if self.tbl_labels[nc] in self.tbl_format:
                        fmt = self.tbl_format[self.tbl_labels[nc]]
                        s = ("{:" + fmt + "}").format(entry)
                    else:
                        s = f"{entry}"

                    item = QTableWidgetItem(s)
                    if nc > 0:
                        item.setTextAlignment(Qt.AlignCenter)
                    else:
                        item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)

                    # Set all columns not editable (unless needed)
                    item.setFlags(item.flags() & ~Qt.ItemIsEditable)

                    # Note, that there is no way to set style sheet for QTableWidgetItem
                    item.setBackground(QBrush(QColor(*rgb_bckg)))

                    self.table.setItem(nr, nc, item)
                else:
                    spin_name = f"{nr}"
                    item = RangeManager(name=spin_name, add_sliders=False, selection_to_range_min=0.0001)
                    item.set_range(0.0, 100.0)  # The range is greater than needed (in keV)
                    item.set_selection(value_low=entry[0], value_high=entry[1])
                    item.setTextColor((0, 0, 0))  # In case of dark theme
                    item.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)

                    self.range_manager_list.append(item)

                    item.selection_changed.connect(self.range_manager_selection_changed)

                    color = (rgb_bckg[0], rgb_bckg[1], rgb_bckg[2])
                    item.setBackground(color)

                    self.table.setCellWidget(nr, nc, item)

            brightness = 220
            if nr % 2:
                rgb_bckg = (255, brightness, brightness)
            else:
                rgb_bckg = (brightness, 255, brightness)

            item = QWidget()
            cb = CheckBoxNamed(name=f"{nr}")
            cb.setChecked(Qt.Checked if range_displayed else Qt.Unchecked)
            self.cb_list.append(cb)
            cb.stateChanged.connect(self.cb_state_changed)

            item_hbox = QHBoxLayout(item)
            item_hbox.addWidget(cb)
            item_hbox.setAlignment(Qt.AlignCenter)
            item_hbox.setContentsMargins(0, 0, 0, 0)
            color_css = f"rgb({rgb_bckg[0]}, {rgb_bckg[1]}, {rgb_bckg[2]})"
            item.setStyleSheet(
                f"QWidget {{ background-color: {color_css}; }} "
                f"QCheckBox {{ color: black; background-color: white }}"
            )
            self.table.setCellWidget(nr, nc + 1, item)

            item = PushButtonNamed("Reset", name=f"{nr}")
            item.clicked.connect(self.pb_default_clicked)
            self.pb_default_list.append(item)
            rgb_bckg = [_ - 35 if (_ < 255) else _ for _ in rgb_bckg]
            color_css = f"rgb({rgb_bckg[0]}, {rgb_bckg[1]}, {rgb_bckg[2]})"
            item.setStyleSheet(f"QPushButton {{ color: black; background-color: {color_css}; }}")
            self.table.setCellWidget(nr, nc + 2, item)

    def _setup_footer(self):

        self.cb_subtract_baseline = QCheckBox("Subtract baseline")
        self.cb_subtract_baseline.setChecked(
            Qt.Checked if self.gpc.get_roi_subtract_background() else Qt.Unchecked
        )
        self.cb_subtract_baseline.toggled.connect(self.cb_subtract_baseline_toggled)

        self.pb_compute_roi = QPushButton("Compute ROIs")
        self.pb_compute_roi.clicked.connect(self.pb_compute_roi_clicked)

        hbox = QHBoxLayout()
        hbox.addWidget(self.cb_subtract_baseline)
        hbox.addStretch(1)
        hbox.addWidget(self.pb_compute_roi)

        return hbox

    def _set_tooltips(self):
        set_tooltip(self.pb_clear, "<b>Clear</b> the list")
        set_tooltip(
            self.pb_use_lines_for_fitting,
            "Copy the contents of <b>the list of emission lines selected for fitting</b> to the list of ROIs",
        )
        set_tooltip(self.le_sel_emission_lines, "The list of <b>emission lines</b> selected for ROI computation.")

        set_tooltip(self.table, "The list of ROIs")

        set_tooltip(
            self.cb_subtract_baseline,
            "<b>Subtract baseline</b> from the pixel spectra before computing ROIs. "
            "Subtracting baseline slows down computations and usually have no benefit. "
            "In most cases it should remain <b>unchecked</b>.",
        )
        set_tooltip(
            self.pb_compute_roi,
            "<b>Run</b> computations of the ROIs. The resulting <b>ROI</b> dataset "
            "may be viewed in <b>XRF Maps</b> tab.",
        )

    def update_widget_state(self, condition=None):
        # Update the state of the menu bar
        state = not self.gui_vars["gui_state"]["running_computations"]
        self.setEnabled(state)

        # Hide the window if required by the program state
        state_file_loaded = self.gui_vars["gui_state"]["state_file_loaded"]
        state_model_exist = self.gui_vars["gui_state"]["state_model_exists"]
        if not state_file_loaded or not state_model_exist:
            self.hide()

        if condition == "tooltips":
            self._set_tooltips()

    def pb_clear_clicked(self):
        self.gpc.clear_roi_element_list()
        self._update_displayed_element_list()
        self._validate_element_list()

    def pb_use_lines_for_fitting_clicked(self):
        self.gpc.load_roi_element_list_from_selected()
        self._update_displayed_element_list()
        self._validate_element_list()

    def le_sel_emission_lines_text_changed(self, text):
        self._validate_element_list(text)

    def le_sel_emission_lines_editing_finished(self):
        text = self.le_sel_emission_lines.text()
        if self._validate_element_list(text):
            self.gpc.set_roi_selected_element_list(text)
            self._update_table()
        else:
            element_list = self.gpc.get_roi_selected_element_list()
            self.le_sel_emission_lines.setText(element_list)

    def cb_subtract_baseline_toggled(self, state):
        self.gpc.set_roi_subtract_background(bool(state))

    def cb_state_changed(self, name, state):
        try:
            nr = int(name)  # Row number
            checked = state == Qt.Checked

            eline = self._table_contents[nr]["eline"]
            self._table_contents[nr]["range_displayed"] = checked
            self.gpc.show_roi(eline, checked)
        except Exception as ex:
            logger.error(f"Failed to process selection change. Exception occurred: {ex}.")

    def _find_spin_box(self, name):
        for item in self.spin_list:
            if item.getName() == name:
                return item
        return None

    def spin_value_changed(self, name, value):
        try:
            nr, side = name.split(",")
            nr = int(nr)
            keys = {"left": "energy_left", "right": "energy_right"}
            side = keys[side]
            eline = self._table_contents[nr]["eline"]
            if self._table_contents[nr][side] == value:
                return
            if side == "energy_left":  # Left boundary
                if value < self._table_contents[nr]["energy_right"]:
                    self._table_contents[nr][side] = value
            else:  # Right boundary
                if value > self._table_contents[nr]["energy_left"]:
                    self._table_contents[nr][side] = value

            # Update plot
            left, right = self._table_contents[nr]["energy_left"], self._table_contents[nr]["energy_right"]
            self.gpc.change_roi(eline, left, right)
        except Exception as ex:
            logger.error(f"Failed to change the ROI. Exception occurred: {ex}.")

    def range_manager_selection_changed(self, left, right, name):
        try:
            nr = int(name)
            eline = self._table_contents[nr]["eline"]
            self.gpc.change_roi(eline, left, right)
        except Exception as ex:
            logger.error(f"Failed to change the ROI. Exception occurred: {ex}.")

    def pb_default_clicked(self, name):
        try:
            nr = int(name)
            eline = self._table_contents[nr]["eline"]
            left = self._table_contents[nr]["energy_left_default"]
            right = self._table_contents[nr]["energy_right_default"]
            self.range_manager_list[nr].set_selection(value_low=left, value_high=right)
            self.gpc.change_roi(eline, left, right)
        except Exception as ex:
            logger.error(f"Failed to change the ROI. Exception occurred: {ex}.")

    def pb_compute_roi_clicked(self):
        def cb():
            try:
                self.gpc.compute_rois()
                success, msg = True, ""
            except Exception as ex:
                success, msg = False, str(ex)

            return {"success": success, "msg": msg}

        self._compute_in_background(cb, self.slot_compute_roi_clicked)

    @Slot(object)
    def slot_compute_roi_clicked(self, result):
        self._recover_after_compute(self.slot_compute_roi_clicked)

        success = result["success"]
        if success:
            self.gui_vars["gui_state"]["state_xrf_map_exists"] = True
        else:
            msg = result["msg"]
            msgbox = QMessageBox(QMessageBox.Critical, "Failed to Compute ROIs", msg, QMessageBox.Ok, parent=self)
            msgbox.exec()

        self.signal_roi_computation_complete.emit()
        self.update_global_state.emit()
        if success:
            self.signal_activate_tab_xrf_maps.emit()

    def _update_displayed_element_list(self):
        element_list = self.gpc.get_roi_selected_element_list()
        self.le_sel_emission_lines.setText(element_list)
        self._validate_element_list()
        self._update_table()

    def _update_table(self):
        table_contents = self.gpc.get_roi_settings()
        self.fill_table(table_contents)

    def _validate_element_list(self, text=None):
        if text is None:
            text = self.le_sel_emission_lines.text()
        el_list = text.split(",")
        el_list = [_.strip() for _ in el_list]
        if el_list == [""]:
            el_list = []
        valid = bool(len(el_list))
        for eline in el_list:
            if self.gpc.get_eline_name_category(eline) != "eline":
                valid = False

        self.le_sel_emission_lines.setValid(valid)
        self.pb_compute_roi.setEnabled(valid)

        return valid

    def _compute_in_background(self, func, slot, *args, **kwargs):
        """
        Run function `func` in a background thread. Send the signal
        `self.computations_complete` once computation is finished.

        Parameters
        ----------
        func: function
            Reference to a function that is supposed to be executed at the background.
            The function return value is passed as a signal parameter once computation is
            complete.
        slot: qtpy.QtCore.Slot or None
            Reference to a slot. If not None, then the signal `self.computation_complete`
            is connected to this slot.
        args, kwargs
            arguments of the function `func`.
        """
        signal_complete = self.computations_complete

        def func_to_run(func, *args, **kwargs):
            class LoadFile(QRunnable):
                def run(self):
                    result_dict = func(*args, **kwargs)
                    signal_complete.emit(result_dict)

            return LoadFile()

        if slot is not None:
            self.computations_complete.connect(slot)
        self.gui_vars["gui_state"]["running_computations"] = True
        self.update_global_state.emit()
        QThreadPool.globalInstance().start(func_to_run(func, *args, **kwargs))

    def _recover_after_compute(self, slot):
        """
        The function should be called after the signal `self.computations_complete` is
        received. The slot should be the same as the one used when calling
        `self.compute_in_background`.
        """
        if slot is not None:
            self.computations_complete.disconnect(slot)
        self.gui_vars["gui_state"]["running_computations"] = False
        self.update_global_state.emit()



import re
import textwrap


class ScanMetadataBase:
    """
    This is a base class. Instances of classes should be created from
    a child class, which has meaningful implementation of the ``_gen_default_descriptions``
    and ``_gen_print_order`` functions.
    """

    def __init__(self):

        # This dictionary contains key-value pairs that represent parameter values
        self._values = {}

        # The following dictionary contains key-description pairs, that represent
        #   user-friendly description of each key, that may be used for printing.
        #   It is not assumed that there is one-to-one match for the keys
        #   in 'values' and 'key_description' dictionaries. The 'key_descriptions'
        #   should contain comprehensive list of keys. It is also assumed that
        #   some keys in 'values' dictionary may not have matching descriptions.
        #   'key_descriptions' dictionary set at the different place in the program.
        #   In principle the descriptions may not be set at all.
        self.descriptions = self._gen_default_descriptions()

        self._print_order = self._gen_default_print_order()

    def _gen_default_print_order(self):
        """
        This function must have meaningful implementation in the child class
        """
        return []

    def _gen_default_descriptions(self):
        """
        This function must have meaningful implementation in the child class
        """
        return {}

    # The following methods override standard operators to enable direct access
    #   to ``self.values`` dictionary

    def __getitem__(self, key):
        return self._values[key]

    def __setitem__(self, key, value):
        self._values[key] = value

    def __delitem__(self, key):
        del self._values[key]

    def __contains__(self, key):
        return key in self._values

    def __iter__(self):
        return self._values.__iter__()

    def keys(self):
        return self._values.keys()

    def values(self):
        return self._values.values()

    def items(self):
        return self._values.items()

    def update(self, source_dict):
        self._values.update(source_dict)

    def get_metadata_dictionary(self):
        """
        Returns reference to metadata dictionary ``self._values``
        """
        return self._values

    def is_metadata_available(self):
        """
        Returns True if at least one metadata key:value pair is available.
        Otherwise returns False.
        """
        return bool(self._values)

    def get_formatted_output(self):
        """
        Returns formatted metadata in the form ready for printing.
        Formatting is performed based on specifications located in ``self._print_order``
        list. Key names are replaced by descriptions from ``self.descriptions`` if
        the available.
        """
        str_out = ""

        printed_keys = set()
        flag_empty_line = False  # Indicates if empty line was just printed

        for ppattern in self._gen_default_print_order():
            # We don't want to print multiple empty lines in a row
            if ppattern == "" and not flag_empty_line:
                str_out += "\n"
                flag_empty_line = True
                continue

            for key, v in self._values.items():
                if re.search(f"^{ppattern}$", key) and (key not in printed_keys):
                    flag_empty_line = False  # Something is getting printed
                    # Mark the entry as printed
                    printed_keys.add(key)
                    # Extracted printable expression for the key
                    s_key = key
                    if key in self.descriptions and self.descriptions[key]:
                        s_key = self.descriptions[key]
                        # Capitalize the first letter
                        s_key = s_key.capitalize()

                    if isinstance(v, str):
                        # Wrap the string if it is too long ('fill' function does not change
                        #   short strings, so call it for every value string)
                        n_indent = len(s_key) + 2
                        text_width = 60
                        indent = " " * n_indent  # Size of the left margin
                        # Create lines with the same indent. Indent width should not be included
                        #   in the text width.
                        val = textwrap.fill(
                            v, width=text_width + n_indent, initial_indent=indent, subsequent_indent=indent
                        )
                        # Now remove spaces at the beginning of the line, since the key will be
                        #   printed instead of the spaces
                        val = val.lstrip()
                    else:
                        val = v

                    # Print the key
                    str_out += f"{s_key}: {val}\n"

        return str_out


class ScanMetadataXRF(ScanMetadataBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _gen_default_print_order(self):
        """
        Generates a list of strings, used to determine printing order of metadata
        The strings in the list are treated as regex strings:
        the symbols ^ and $ are added at the beginning and the end of each string.
        Empty string "" means that the empty string is inserted in the printout.
        The metadata entries are never repeated in the printout, so in each group
        the patterns are specified from more specific to more general.
        """
        print_order = [
            "scan_id",
            "scan_uid",
            "scan_instrument_name",
            "scan_instrument_id",
            "scan_time_start",
            "scan_time_start_utc",
            "scan_time_stop",
            "scan_time_stop_utc",
            "scan.*",
            "",
            "sample_name",
            "sample.*",
            "",
            "param_type",
            "param_input",
            "param_shape",
            "param_snake",
            "param_dwell",
            "param_theta",
            "param_theta_units",
            "param_delta",
            "param_delta_units",
            "param_fast_axis",
            "param_fast_axis_units",
            "param_slow_axis",
            "param_slow_axis_units",
            "param.*",
            "",
            "proposal_num",
            "proposal_title",
            "proposal.*",
            "",
            "experiment.*",
            "",
            "instrument.*",
            "",
            "(?!file).*",
            "",
            "file_format",
            "file_format_version",
            "file_software",
            "file_software_version",
            "file.*",
        ]
        return print_order

    def _gen_default_descriptions(self):
        descriptions = {
            # The descriptions are not capitalized. They can be capitalized
            #   before printing if needed.
            "scan_id": "scan ID",
            "scan_uid": "scan Unique ID",
            "scan_time_start": "start time",
            "scan_time_stop": "stop time",
            "scan_time_start_utc": "start time (UTC)",
            "scan_time_stop_utc": "stop time (UTC)",
            "scan_instrument_id": "beamline ID",
            "scan_instrument_name": "beamline name",
            "scan_exit_status": "exit status",
            "instrument_mono_incident_energy": "incident energy",
            "instrument_beam_current": "beam current",
            "instrument_detectors": "detectors",
            "sample_name": "sample name",
            "experiment_plan_name": "plan name",
            "experiment_plan_type": "plan type",
            "experiment_fast_axis": "scan fast axis",
            "experiment_slow_axis": "scan slow axis",
            "proposal_num": "proposal #",
            "proposal_title": "proposal title",
            "proposal_PI_lastname": "PI last name",
            "proposal_saf_num": "proposal SAF #",
            "proposal_cycle": "cycle",
            "file_created_time": "file creation time",
            "file_format": "file format",
            "file_format_version": "file format version",
            "file_software": "software",
            "file_software_version": "version",
            "file_type": "file type",
            "param_type": "plan type",
            "param_input": "plan input parameters",
            "param_shape": "scan shape",
            "param_snake": "snaking",
            "param_dwell": "dwell time",
            "param_theta": "angle theta",
            "param_theta_units": "angle theta (units)",
            "param_delta": "delta",
            "param_delta_units": "delta (units)",
            "param_fast_axis": "fast axis",
            "param_fast_axis_units": "fast axis (units)",
            "param_slow_axis": "slow axis",
            "param_slow_axis_units": "slow axis (units)",
        }

        return descriptions

    def is_mono_incident_energy_available(self):
        """
        Returns True if data on monochromator incident energy is available.
        Otherwise returns False
        """
        return "instrument_mono_incident_energy" in self._values

    def get_mono_incident_energy(self):
        """
        Returns the value of the incident energy. Incident energy is an important
        parameter used in processing, so a separate function is created to fetch it.
        """
        return self._values["instrument_mono_incident_energy"]



import textwrap

from qtpy.QtWidgets import (
    QPushButton,
    QHBoxLayout,
    QVBoxLayout,
    QGroupBox,
    QCheckBox,
    QLabel,
    QComboBox,
    QFileDialog,
    QRadioButton,
    QButtonGroup,
    QTableWidget,
    QTableWidgetItem,
    QHeaderView,
    QWidget,
    QScrollArea,
    QTabWidget,
    QFrame,
    QMessageBox,
)
from qtpy.QtGui import QBrush, QColor, QDoubleValidator
from qtpy.QtCore import Qt, Slot, Signal

from .useful_widgets import get_background_css, SecondaryWindow, set_tooltip, LineEditExtended
from .dlg_view_calib_standard import DialogViewCalibStandard

import logging

logger = logging.getLogger(__name__)


class WndLoadQuantitativeCalibration(SecondaryWindow):

    signal_quantitative_calibration_changed = Signal()

    def __init__(self, *, gpc, gui_vars):
        super().__init__()

        # Global processing classes
        self.gpc = gpc
        # Global GUI variables (used for control of GUI state)
        self.gui_vars = gui_vars

        self.initialize()

    def initialize(self):
        self.table_header_display_names = False

        self.setWindowTitle("PyXRF: Load Quantitative Calibration")
        self.setMinimumWidth(750)
        self.setMinimumHeight(400)
        self.resize(750, 600)

        self.pb_load_calib = QPushButton("Load Calibration ...")
        self.pb_load_calib.clicked.connect(self.pb_load_calib_clicked)

        self._changes_exist = False
        self._auto_update = True
        self.cb_auto_update = QCheckBox("Auto")
        self.cb_auto_update.setCheckState(Qt.Checked if self._auto_update else Qt.Unchecked)
        self.cb_auto_update.stateChanged.connect(self.cb_auto_update_state_changed)

        self.pb_update_plots = QPushButton("Update Plots")
        self.pb_update_plots.clicked.connect(self.pb_update_plots_clicked)

        self.grp_current_scan = QGroupBox("Parameters of Currently Processed Scan")

        self._distance_to_sample = 0.0
        self.le_distance_to_sample = LineEditExtended()
        le_dist_validator = QDoubleValidator()
        le_dist_validator.setBottom(0)
        self.le_distance_to_sample.setValidator(le_dist_validator)
        self._set_distance_to_sample()
        self.le_distance_to_sample.editingFinished.connect(self.le_distance_to_sample_editing_finished)
        self.le_distance_to_sample.focusOut.connect(self.le_distance_to_sample_focus_out)

        hbox = QHBoxLayout()
        hbox.addWidget(QLabel("Distance-to-sample:"))
        hbox.addWidget(self.le_distance_to_sample)
        hbox.addStretch(1)
        self.grp_current_scan.setLayout(hbox)

        self.eline_rb_exclusive = []  # Holds the list of groups of exclusive radio buttons
        self._setup_tab_widget()

        vbox = QVBoxLayout()

        hbox = QHBoxLayout()
        hbox.addWidget(self.pb_load_calib)
        hbox.addStretch(1)
        hbox.addWidget(self.cb_auto_update)
        hbox.addWidget(self.pb_update_plots)
        vbox.addLayout(hbox)

        vbox.addWidget(self.tab_widget)

        vbox.addWidget(self.grp_current_scan)

        self.setLayout(vbox)

        # Display data
        self.update_all_data()

        self._set_tooltips()

    def _setup_tab_widget(self):

        self.tab_widget = QTabWidget()
        self.loaded_standards = QWidget()
        # self.display_loaded_standards()
        self.scroll = QScrollArea()
        self.scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.scroll.setWidget(self.loaded_standards)
        self.tab_widget.addTab(self.scroll, "Loaded Standards")

        self.combo_set_table_header = QComboBox()
        self.combo_set_table_header.addItems(["Standard Serial #", "Standard Name"])
        self.combo_set_table_header.currentIndexChanged.connect(self.combo_set_table_header_index_changed)

        vbox = QVBoxLayout()
        vbox.addSpacing(5)
        hbox = QHBoxLayout()
        hbox.addWidget(QLabel("Display in table header:"))
        hbox.addWidget(self.combo_set_table_header)
        hbox.addStretch(1)
        vbox.addLayout(hbox)
        self.table = QTableWidget()
        self.table.verticalHeader().hide()
        self.table.setSelectionMode(QTableWidget.NoSelection)
        self.table.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.table.horizontalHeader().setMinimumSectionSize(150)
        vbox.addWidget(self.table)

        self.table.setStyleSheet("QTableWidget::item{color: black;}")

        frame = QFrame()
        vbox.setContentsMargins(0, 0, 0, 0)
        frame.setLayout(vbox)

        self.tab_widget.addTab(frame, "Selected Emission Lines")

    def display_loaded_standards(self):
        calib_data = self.gpc.get_quant_calibration_data()
        calib_settings = self.gpc.get_quant_calibration_settings()

        # Create the new widget (this deletes the old widget)
        self.loaded_standards = QWidget()
        self.loaded_standards.setMinimumWidth(700)

        # Also delete references to all components
        self.frames_calib_data = []
        self.pbs_view = []
        self.pbs_remove = []

        # All 'View' buttons are added to the group in order to be connected to the same slot
        self.group_view = QButtonGroup()
        self.group_view.setExclusive(False)
        self.group_view.buttonClicked.connect(self.pb_view_clicked)
        # The same for the 'Remove' buttons
        self.group_remove = QButtonGroup()
        self.group_remove.setExclusive(False)
        self.group_remove.buttonClicked.connect(self.pb_remove_clicked)

        vbox = QVBoxLayout()

        class _LabelBlack(QLabel):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.setStyleSheet("color: black")

        for cdata, csettings in zip(calib_data, calib_settings):
            frame = QFrame()
            frame.setFrameStyle(QFrame.StyledPanel)
            frame.setStyleSheet(get_background_css((200, 255, 200), widget="QFrame"))

            _vbox = QVBoxLayout()

            name = cdata["name"]  # Standard name (can be arbitrary string
            # If name is long, then print it in a separate line
            _name_is_long = len(name) > 30

            pb_view = QPushButton("View ...")
            self.group_view.addButton(pb_view)
            pb_remove = QPushButton("Remove")
            self.group_remove.addButton(pb_remove)

            # Row 1: serial, name
            serial = cdata["serial"]
            _hbox = QHBoxLayout()
            _hbox.addWidget(_LabelBlack(f"<b>Standard</b> #{serial}"))
            if not _name_is_long:
                _hbox.addWidget(_LabelBlack(f"'{name}'"))
            _hbox.addStretch(1)
            _hbox.addWidget(pb_view)
            _hbox.addWidget(pb_remove)
            _vbox.addLayout(_hbox)

            # Optional row
            if _name_is_long:
                # Wrap name if it is extemely long
                name = textwrap.fill(name, width=80)
                _hbox = QHBoxLayout()
                _hbox.addWidget(_LabelBlack("<b>Name:</b> "), 0, Qt.AlignTop)
                _hbox.addWidget(_LabelBlack(name), 0, Qt.AlignTop)
                _hbox.addStretch(1)
                _vbox.addLayout(_hbox)

            # Row 2: description
            description = textwrap.fill(cdata["description"], width=80)
            _hbox = QHBoxLayout()
            _hbox.addWidget(_LabelBlack("<b>Description:</b>"), 0, Qt.AlignTop)
            _hbox.addWidget(_LabelBlack(f"{description}"), 0, Qt.AlignTop)
            _hbox.addStretch(1)
            _vbox.addLayout(_hbox)

            # Row 3:
            incident_energy = cdata["incident_energy"]
            scaler = cdata["scaler_name"]
            detector_channel = cdata["detector_channel"]
            distance_to_sample = cdata["distance_to_sample"]
            _hbox = QHBoxLayout()
            _hbox.addWidget(_LabelBlack(f"<b>Incident energy, keV:</b> {incident_energy}"))
            _hbox.addWidget(_LabelBlack(f"  <b>Scaler:</b> {scaler}"))
            _hbox.addWidget(_LabelBlack(f"  <b>Detector channel:</b> {detector_channel}"))
            _hbox.addWidget(_LabelBlack(f"  <b>Distance-to-sample:</b> {distance_to_sample}"))
            _hbox.addStretch(1)
            _vbox.addLayout(_hbox)

            # Row 4: file name
            fln = textwrap.fill(csettings["file_path"], width=80)
            _hbox = QHBoxLayout()
            _hbox.addWidget(_LabelBlack("<b>Source file:</b>"), 0, Qt.AlignTop)
            _hbox.addWidget(_LabelBlack(fln), 0, Qt.AlignTop)
            _hbox.addStretch(1)
            _vbox.addLayout(_hbox)

            frame.setLayout(_vbox)

            # Now the group box is added to the upper level layout
            vbox.addWidget(frame)
            vbox.addSpacing(5)
            self.frames_calib_data.append(frame)
            self.pbs_view.append(pb_view)
            self.pbs_remove.append(pb_remove)

        # Add the layout to the widget
        self.loaded_standards.setLayout(vbox)
        # ... and put the widget inside the scroll area. This will update the
        # contents of the scroll area.
        self.scroll.setWidget(self.loaded_standards)

    def display_table_header(self):
        calib_data = self.gpc.get_quant_calibration_data()
        header_by_name = self.table_header_display_names

        tbl_labels = ["Lines"]
        for n, cdata in enumerate(calib_data):
            if header_by_name:
                txt = cdata["name"]
            else:
                txt = cdata["serial"]
            txt = textwrap.fill(txt, width=20)
            tbl_labels.append(txt)

        self.table.setHorizontalHeaderLabels(tbl_labels)

    def display_standard_selection_table(self):
        calib_data = self.gpc.get_quant_calibration_data()
        self._quant_file_paths = self.gpc.get_quant_calibration_file_path_list()

        brightness = 220
        table_colors = [(255, brightness, brightness), (brightness, 255, brightness)]

        # Disconnect all radio button signals before clearing the table
        for bgroup in self.eline_rb_exclusive:
            bgroup.buttonToggled.disconnect(self.rb_selection_toggled)

        # This list will hold radio button groups for horizontal rows
        #   Those are exclusive groups. They are not going to be
        #   used directly, but they must be kept alive in order
        #   for the radiobuttons to work properly. Most of the groups
        #   will contain only 1 radiobutton, which will always remain checked.
        self.eline_rb_exclusive = []
        # The following list will contain the list of radio buttons for each
        #   row. If there is no radiobutton in a position, then the element is
        #   set to None.
        # N rows: the number of emission lines, N cols: the number of standards
        self.eline_rb_lists = []

        self.table.clear()

        if not calib_data:
            self.table.setRowCount(0)
            self.table.setColumnCount(0)
        else:
            # Create the sorted list of available element lines
            line_set = set()
            for cdata in calib_data:
                ks = list(cdata["element_lines"].keys())
                line_set.update(list(ks))
            self.eline_list = list(line_set)
            self.eline_list.sort()

            for n in range(len(self.eline_list)):
                self.eline_rb_exclusive.append(QButtonGroup())
                self.eline_rb_lists.append([None] * len(calib_data))

            self.table.setColumnCount(len(calib_data) + 1)
            self.table.setRowCount(len(self.eline_list))
            self.display_table_header()

            for n, eline in enumerate(self.eline_list):

                rgb = table_colors[n % 2]

                item = QTableWidgetItem(eline)
                item.setTextAlignment(Qt.AlignCenter)
                item.setFlags(item.flags() & ~Qt.ItemIsEditable)
                item.setBackground(QBrush(QColor(*rgb)))
                self.table.setItem(n, 0, item)

                for ns, cdata in enumerate(calib_data):
                    q_file_path = self._quant_file_paths[ns]  # Used to identify standard
                    if eline in cdata["element_lines"]:
                        rb = QRadioButton()
                        if self.gpc.get_quant_calibration_is_eline_selected(eline, q_file_path):
                            rb.setChecked(True)

                        rb.setStyleSheet("color: black")

                        self.eline_rb_lists[n][ns] = rb
                        # self.eline_rb_by_standard[ns].addButton(rb)
                        self.eline_rb_exclusive[n].addButton(rb)

                        item = QWidget()
                        item_hbox = QHBoxLayout(item)
                        item_hbox.addWidget(rb)
                        item_hbox.setAlignment(Qt.AlignCenter)
                        item_hbox.setContentsMargins(0, 0, 0, 0)

                        item.setStyleSheet(get_background_css(rgb))

                        # Generate tooltip
                        density = cdata["element_lines"][eline]["density"]
                        fluorescence = cdata["element_lines"][eline]["fluorescence"]
                        ttip = f"Fluorescence (F): {fluorescence:12g}\nDensity (D): {density:12g}\n"
                        # Avoid very small values of density (probably zero)
                        if abs(density) > 1e-30:
                            ttip += f"F/D: {fluorescence/density:12g}"

                        item.setToolTip(ttip)

                        self.table.setCellWidget(n, ns + 1, item)
                    else:
                        # There is no radio button, but we still need to fill the cell
                        item = QTableWidgetItem("")
                        item.setFlags(item.flags() & ~Qt.ItemIsEditable)
                        item.setBackground(QBrush(QColor(*rgb)))
                        self.table.setItem(n, ns + 1, item)

            # Now the table is set (specifically radio buttons).
            # So we can connect the button groups with the event processing function
            for bgroup in self.eline_rb_exclusive:
                bgroup.buttonToggled.connect(self.rb_selection_toggled)

    @Slot()
    def update_all_data(self):
        self.display_loaded_standards()
        self.display_standard_selection_table()
        self._set_distance_to_sample()

    def _set_distance_to_sample(self):
        """Set 'le_distance_to_sample` without updating maps"""
        distance_to_sample = self.gpc.get_quant_calibration_distance_to_sample()
        if distance_to_sample is None:
            distance_to_sample = 0.0
        self._distance_to_sample = distance_to_sample
        self._set_le_distance_to_sample(distance_to_sample)

    def _set_tooltips(self):
        set_tooltip(self.pb_load_calib, "Load <b>calibration data</b> from JSON file.")
        set_tooltip(
            self.cb_auto_update,
            "Automatically <b>update the plots</b> when changes are made. "
            "If unchecked, then button <b>Update Plots</b> must be pressed "
            "to update the plots. Automatic update is often undesirable "
            "when large maps are displayed and multiple changes to parameters "
            "are made.",
        )
        set_tooltip(self.pb_update_plots, "<b>Update plots</b> based on currently selected parameters.")
        set_tooltip(
            self.le_distance_to_sample,
            "Distance between <b>the sample and the detector</b>. The ratio between of the distances "
            "during calibration and measurement is used to scale computed concentrations. "
            "If distance-to-sample is 0 for calibration or measurement, then no scaling is performed.",
        )
        set_tooltip(
            self.combo_set_table_header,
            "Use <b>Serial Number</b> or <b>Name</b> of the calibration standard in the header of the table",
        )
        set_tooltip(
            self.table,
            "Use Radio Buttons to select the <b>source of calibration data</b> for each emission line. "
            "This feature is needed if multiple loaded calibration files have data on the same "
            "emission line.",
        )

    def update_widget_state(self, condition=None):
        # Update the state of the menu bar
        state = not self.gui_vars["gui_state"]["running_computations"]
        self.setEnabled(state)

        # Hide the window if required by the program state
        state_xrf_map_exists = self.gui_vars["gui_state"]["state_xrf_map_exists"]
        if not state_xrf_map_exists:
            self.hide()

        if condition == "tooltips":
            self._set_tooltips()

    def cb_auto_update_state_changed(self, state):
        self._auto_update = state
        self.pb_update_plots.setEnabled(not state)
        # If changes were made, apply the changes while switching to 'auto' mode
        if state and self._changes_exist:
            self._update_maps_auto()

    def pb_update_plots_clicked(self):
        self._update_maps()

    def pb_load_calib_clicked(self):
        current_dir = self.gpc.get_current_working_directory()
        file_name = QFileDialog.getOpenFileName(
            self, "Select File with Quantitative Calibration Data", current_dir, "JSON (*.json);; All (*)"
        )
        file_name = file_name[0]
        if file_name:
            try:
                logger.debug(f"Loading quantitative calibration from file: '{file_name}'")
                self.gpc.load_quantitative_calibration_data(file_name)
                self.update_all_data()
                self._update_maps_auto()
            except Exception:
                msg = "The selected JSON file has incorrect format. Select a different file."
                msgbox = QMessageBox(QMessageBox.Critical, "Data Loading Error", msg, QMessageBox.Ok, parent=self)
                msgbox.exec()

    def pb_view_clicked(self, button):
        try:
            n_standard = self.pbs_view.index(button)
            calib_settings = self.gpc.get_quant_calibration_settings()
            file_path = calib_settings[n_standard]["file_path"]
            calib_preview = self.gpc.get_quant_calibration_text_preview(file_path)
            dlg = DialogViewCalibStandard(None, file_path=file_path, calib_preview=calib_preview)
            dlg.exec()
        except ValueError:
            logger.error("'View' button was pressed, but not found in the list of buttons")

    def pb_remove_clicked(self, button):
        try:
            n_standard = self.pbs_remove.index(button)
            calib_settings = self.gpc.get_quant_calibration_settings()
            file_path = calib_settings[n_standard]["file_path"]
            self.gpc.quant_calibration_remove_entry(file_path)
            self.update_all_data()
            self._update_maps_auto()
        except ValueError:
            logger.error("'Remove' button was pressed, but not found in the list of buttons")

    def rb_selection_toggled(self, button, checked):
        if checked:
            # Find the button in 2D list 'self.eline_rb_lists'
            button_found = False
            for nr, rb_list in enumerate(self.eline_rb_lists):
                try:
                    nc = rb_list.index(button)
                    button_found = True
                    break
                except ValueError:
                    pass

            if button_found:
                eline = self.eline_list[nr]
                n_standard = nc
                file_path = self._quant_file_paths[n_standard]
                self.gpc.set_quant_calibration_select_eline(eline, file_path)
                self._update_maps_auto()
            else:
                # This should never happen
                logger.error("Selection radio button was pressed, but not found in the list")

    def combo_set_table_header_index_changed(self, index):
        self.table_header_display_names = bool(index)
        self.display_table_header()

    def le_distance_to_sample_editing_finished(self):
        distance_to_sample = float(self.le_distance_to_sample.text())
        if distance_to_sample != self._distance_to_sample:
            self._distance_to_sample = distance_to_sample
            self.gpc.set_quant_calibration_distance_to_sample(distance_to_sample)
            self._update_maps_auto()

    def le_distance_to_sample_focus_out(self):
        try:
            float(self.le_distance_to_sample.text())
        except ValueError:
            # If the text can not be interpreted to float, then replace the text with the old value
            self._set_le_distance_to_sample(self._distance_to_sample)

    def _set_le_distance_to_sample(self, distance_to_sample):
        self.le_distance_to_sample.setText(f"{distance_to_sample:.12g}")

    def _update_maps_auto(self):
        """Update maps only if 'auto' update is ON. Used as a 'filter'
        to prevent extra plot updates."""
        self._changes_exist = True
        if self._auto_update:
            self._update_maps()

    def _update_maps(self):
        """Upload the selections (limit table) and update plot"""
        self._changes_exist = False
        self._redraw_maps()
        # Emit signal only after the maps are redrawn. This should change
        #   ranges in the respective controls for the plots
        self.signal_quantitative_calibration_changed.emit()

    def _redraw_maps(self):
        # We don't emit any signals here, but we don't really need to.
        logger.debug("Redrawing RGB XRF Maps")
        self.gpc.compute_map_ranges()
        self.gpc.redraw_maps()
        self.gpc.compute_rgb_map_ranges()
        self.gpc.redraw_rgb_maps()



import os
import pytest
from pyxrf.gui_module.tab_wd_load_data import DialogLoadMask
from PyQt5.QtWidgets import QFileDialog
from PyQt5.QtCore import Qt


# =====================================================================
#                 class DialogLoadMask (QDialogBox)


def test_DialogLoadMask_1(qtbot):
    """Set/Read values from ROI widget group of DialogLoadMask object"""
    dlg = DialogLoadMask()
    qtbot.addWidget(dlg)
    dlg.show()

    # Set image size
    n_rows, n_columns = 15, 17
    dlg.set_image_size(n_rows=n_rows, n_columns=n_columns)
    assert f"{n_rows} rows, {n_columns} columns" in dlg.label_map_size.text(), "Map size was not set correctly"

    def _check_roi_widgets(expected_values):
        l_edit_widgets = (dlg.le_roi_start_row, dlg.le_roi_start_col, dlg.le_roi_end_row, dlg.le_roi_end_col)
        for n, v in enumerate(expected_values):
            if n < 2:  # First 2 entries (are incremented when displayed)
                v += 1
            s = f"{v}" if v > 0 else ""
            assert l_edit_widgets[n].text() == s, "Line Edit text does not match the expected text"

    # Check if the default values were set correctly when image size was set
    #   (try to activate/deactivate the ROI selection group box)
    _check_roi_widgets([0, 0, n_rows, n_columns])  # Default values
    dlg.set_roi_active(True)  # Activate ROI group
    _check_roi_widgets([0, 0, n_rows, n_columns])  # Default values
    dlg.set_roi_active(False)
    _check_roi_widgets([0, 0, n_rows, n_columns])  # Default values

    # Set ROI. Initially the ROI group is inactive, so no text is displayed
    roi = (2, 3, 14, 15)
    dlg.set_roi(row_start=roi[0], column_start=roi[1], row_end=roi[2], column_end=roi[3])
    assert not dlg.gb_roi.isChecked(), "ROI group is checked, while it must be unchecked"
    assert dlg.get_roi_active() is False, "ROI 'enable' status was not read correctly"
    _check_roi_widgets(roi)
    # Activate ROI group
    dlg.set_roi_active(True)
    assert dlg.gb_roi.isChecked(), "ROI group is unchecked, while it must be checked"
    assert dlg.get_roi_active() is True, "ROI 'enable' status was not read correctly"
    _check_roi_widgets(roi)

    # Set the values outside the range. The values must be clipped
    roi = (0, 0, 50, 50)
    dlg.set_roi(row_start=roi[0], column_start=roi[1], row_end=roi[2], column_end=roi[3])
    assert dlg.get_roi() == (roi[0], roi[1], n_rows, n_columns), "Indexes were not clipped correctly"

    dlg.close()


def test_DialogLoadMask_2(qtbot):
    """
    DialogLoadMask - ROI selection group.
    In this test, the ROI selection is modified using keyboard clicks.
    Checked features: initalization, enabling/disabling ROI selection group, editing the ROI entries,
    synchronization of entered and stored values, showing/hiding the displayed values as
    the group is enabled/disabled.
    """
    dlg = DialogLoadMask()
    qtbot.addWidget(dlg)
    dlg.show()

    def _check_roi_widgets(expected_values):
        l_edit_widgets = (dlg.le_roi_start_row, dlg.le_roi_start_col, dlg.le_roi_end_row, dlg.le_roi_end_col)
        for n, v in enumerate(expected_values):
            if n < 2:  # First 2 entries (are incremented when displayed)
                v += 1
            s = f"{v}" if v > 0 else ""
            assert l_edit_widgets[n].text() == s, "Line Edit text does not match the expected text"

    # Set image size
    n_rows, n_columns = 25, 37
    dlg.set_image_size(n_rows=n_rows, n_columns=n_columns)

    assert not dlg.gb_roi.isChecked(), "ROI selection group must be active"
    _check_roi_widgets([0, 0, n_rows, n_columns])
    dlg.gb_roi.setChecked(True)  # There seems to be no way to change the state of the group box by clicking
    assert dlg.gb_roi.isChecked(), "Failed to activate ROI group"
    _check_roi_widgets([0, 0, n_rows, n_columns])

    # Use keyboard to enter values in each box
    def _enter_value(edit_box, value):
        qtbot.mouseDClick(edit_box, Qt.LeftButton)
        qtbot.keyClicks(edit_box, f"{value}")
        qtbot.keyClick(edit_box, Qt.Key_Enter)

    # Enter a new ROI using keyboard
    sel = (11, 8, 23, 19)
    _enter_value(dlg.le_roi_start_row, sel[0] + 1)
    _enter_value(dlg.le_roi_start_col, sel[1] + 1)
    _enter_value(dlg.le_roi_end_row, sel[2])
    _enter_value(dlg.le_roi_end_col, sel[3])
    _check_roi_widgets(sel)
    # Now check if the values were set correctly
    assert dlg.get_roi() == sel, "ROI was incorrectly set"

    # Disable and then enable the ROI group again, the entered values should remain
    dlg.gb_roi.setChecked(False)
    assert dlg.get_roi_active() is False, "ROI group must be inactive"
    _check_roi_widgets(sel)  # Values must be hidden now
    assert dlg.get_roi() == sel, "Unexpected ROI selection values"
    dlg.gb_roi.setChecked(True)
    assert dlg.get_roi_active() is True, "ROI group must be active"
    _check_roi_widgets(sel)  # Values must be displayed
    assert dlg.get_roi() == sel, "Unexpected ROI selection values"


@pytest.mark.parametrize("use_keys_and_mouse", [False, True])
def test_DialogLoadMask_3(qtbot, use_keys_and_mouse):
    """
    DialogLoadMask - ROI selection group.
    Check if validation of dialog controls works as expected.
    In this test the values are set using dialog box functions, not using mouse/keyboard input.
    """

    dlg = DialogLoadMask()
    qtbot.addWidget(dlg)
    dlg.show()

    # Set image size
    n_rows, n_columns = 25, 37
    dlg.set_image_size(n_rows=n_rows, n_columns=n_columns)

    def _set_roi_group_active(status):
        if use_keys_and_mouse:
            dlg.set_roi_active(status)
        else:
            dlg.set_roi_active(status)

    def _enter_value(edit_box, value):
        """Use keyboard to enter values line edit box"""
        qtbot.mouseDClick(edit_box, Qt.LeftButton)
        qtbot.keyClicks(edit_box, f"{value}")
        qtbot.keyClick(edit_box, Qt.Key_Enter)

    def _set_roi(roi):
        if use_keys_and_mouse:
            dlg.set_roi(row_start=roi[0], column_start=roi[1], row_end=roi[2], column_end=roi[3])
        else:
            _enter_value(dlg.le_roi_start_row, roi[0])
            _enter_value(dlg.le_roi_start_col, roi[1])
            _enter_value(dlg.le_roi_end_row, roi[2])
            _enter_value(dlg.le_roi_end_col, roi[3])

    def _check_controls(status):
        assert dlg.le_roi_start_row.isValid() is status[0], "Invalid status: ROI control - row start"
        assert dlg.le_roi_start_col.isValid() is status[1], "Invalid status: ROI control - column start"
        assert dlg.le_roi_end_row.isValid() is status[2], "Invalid status: ROI control - row end"
        assert dlg.le_roi_end_col.isValid() is status[3], "Invalid status: ROI control - column end"
        assert dlg.button_ok.isEnabled() is status[4], "Invalid status: Button Ok"

    # Activate ROI selection group
    _set_roi_group_active(True)

    # Set valid selection
    sel = (11, 8, 23, 19)
    _set_roi(sel)
    _check_controls([True, True, True, True, True])

    # Set invalid selection for rows
    sel = (23, 8, 4, 19)
    _set_roi(sel)
    _check_controls([False, True, False, True, False])

    # Set invalid selection for columns
    sel = (11, 19, 23, 5)
    _set_roi(sel)
    _check_controls([True, False, True, False, False])

    # Set invalid selection for both
    sel = (23, 19, 11, 5)
    _set_roi(sel)
    _check_controls([False, False, False, False, False])

    # Deactivate ROI selection group
    _set_roi_group_active(False)
    # Check only the status of Ok button
    assert dlg.button_ok.isEnabled() is True, "Invalid status: Button Ok"


@pytest.mark.parametrize("use_keys_and_mouse", [False, True])
def test_DialogLoadMask_4(qtbot, use_keys_and_mouse):
    """
    DialogLoadMask - ROI selection group.
    Check if validation of dialog controls works as expected.
    In this test the values are set using dialog box functions, not using mouse/keyboard input.
    """

    dlg = DialogLoadMask()
    qtbot.addWidget(dlg)
    dlg.show()

    # Set image size
    n_rows, n_columns = 25, 37
    dlg.set_image_size(n_rows=n_rows, n_columns=n_columns)

    def _activate_mask_selection_group(state):
        if use_keys_and_mouse:
            dlg.gb_mask.setChecked(state)
        else:
            dlg.set_mask_file_active(state)
        assert dlg.get_mask_file_active() == state, "Mask selection group status is incorrect"
        assert dlg.gb_mask.isChecked() == state, "Mask selection group check status is incorrect"

    assert dlg.le_load_mask.isValid() is True, "Validation status of the mask file name is incorrect"
    _activate_mask_selection_group(True)
    assert dlg.le_load_mask.isValid() is False, "Validation status of the mask file name is incorrect"
    _activate_mask_selection_group(False)
    assert dlg.le_load_mask.isValid() is True, "Validation status of the mask file name is incorrect"

    fpath = "/file/path/file_name.bin"  # File path doesn't have to be real

    # Set file name during 'inactive' state
    dlg.set_mask_file_path(fpath)
    assert dlg.le_load_mask.isValid() is True, "Validation status of the mask file name is incorrect"
    _activate_mask_selection_group(True)
    assert dlg.le_load_mask.isValid() is True, "Validation status of the mask file name is incorrect"
    _activate_mask_selection_group(False)
    assert dlg.le_load_mask.isValid() is True, "Validation status of the mask file name is incorrect"

    # Set file name during 'inactive' state
    dlg.set_mask_file_path("")  # Clear the file path
    assert dlg.le_load_mask.isValid() is True, "Validation status of the mask file name is incorrect"
    _activate_mask_selection_group(True)
    assert dlg.le_load_mask.isValid() is False, "Validation status of the mask file name is incorrect"
    dlg.set_mask_file_path(fpath)
    assert dlg.le_load_mask.isValid() is True, "Validation status of the mask file name is incorrect"
    _activate_mask_selection_group(False)
    assert dlg.le_load_mask.isValid() is True, "Validation status of the mask file name is incorrect"


def test_DialogLoadMask_5(qtbot):
    """Setting home directory"""

    dlg = DialogLoadMask()
    qtbot.addWidget(dlg)
    dlg.show()

    # Set image size
    n_rows, n_columns = 25, 37
    dlg.set_image_size(n_rows=n_rows, n_columns=n_columns)

    assert dlg._compute_home_directory() == ".", "Home directory is computed incorrectly"
    d_name = os.path.join("file", "directory")
    d_name = os.path.abspath(d_name)
    dlg.set_default_directory(d_name)
    assert dlg.get_default_directory() == d_name, "Returned default directory is incorrect"
    assert dlg._compute_home_directory() == d_name, "Home directory is computed incorrectly"

    d_name2 = os.path.join("file", "directory")
    d_name2 = os.path.abspath(d_name2)
    dlg.set_default_directory(d_name2)
    f_path = os.path.join(d_name2, "file_name.bin")
    dlg.set_mask_file_path(f_path)
    assert dlg._compute_home_directory() == d_name2, "Home directory is computed incorrectly"

    # Now clear the default directory
    dlg.set_default_directory(d_name)
    assert dlg._compute_home_directory() == d_name2, "Home directory is computed incorrectly"


def test_DialogLoadMask_6(qtbot, monkeypatch):
    """Setting home directory"""

    dlg = DialogLoadMask()
    dlg.show()
    qtbot.addWidget(dlg)
    qtbot.waitForWindowShown(dlg)

    # Set image size
    n_rows, n_columns = 25, 37
    dlg.set_image_size(n_rows=n_rows, n_columns=n_columns)

    d_name = os.path.join("file", "directory")
    d_name = os.path.abspath(d_name)
    f_name = "file_name.bin"
    d_path = os.path.join(d_name, f_name)
    dlg.set_default_directory(d_name)

    # Activate the group
    dlg.gb_mask.setChecked(True)

    monkeypatch.setattr(QFileDialog, "getOpenFileName", lambda *args: [""])
    qtbot.mouseClick(dlg.pb_load_mask, Qt.LeftButton)
    assert dlg.le_load_mask.isValid() is False, "Incorrect state of the line edit widget"
    assert dlg.le_load_mask.text() == dlg._le_load_mask_default_text, "Line Edit text is set incorrectly"
    assert dlg.get_mask_file_path() == "", "Unexpected returned path for the mask file"

    monkeypatch.setattr(QFileDialog, "getOpenFileName", lambda *args: [d_path])
    qtbot.mouseClick(dlg.pb_load_mask, Qt.LeftButton)
    assert dlg.le_load_mask.isValid() is True, "Incorrect state of the line edit widget"
    assert dlg.le_load_mask.text() == d_path, "Line Edit text is set incorrectly"
    assert dlg.get_mask_file_path() == d_path, "Unexpected returned path for the mask file"



from qtpy.QtWidgets import QWidget, QSizePolicy


class FormBaseWidget(QWidget):
    def __init__(self):
        super().__init__()
        self.setMaximumWidth(400)
        self.setMinimumWidth(400)

        sp = self.sizePolicy()
        sp.setVerticalPolicy(QSizePolicy.Minimum)
        self.setSizePolicy(sp)
